/*
 * Copyright (c) 2022 Arm Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#if defined(ARM_COMPUTE_ENABLE_SME2)

#include <algorithm>
#include <cstddef>
#include "arm_gemm.hpp"

using arm_gemm::Requantize32;

namespace arm_conv {
namespace depthwise {

void sme2_s8q_planar_5x5_s1_4rows_dot_za_impl(
  const int8_t *inptr,
  size_t ld_in_row,
  size_t ld_in_col,
  size_t ld_in_vl,
  unsigned int pad_top,
  unsigned int valid_input_rows,
  unsigned int pad_left,
  unsigned int valid_input_cols,
  const int8_t *weights,
  int8_t **outptrs,
  const size_t *outlds,
  const size_t *outvllds,
  unsigned int output_cols,
  unsigned int start_channel,
  unsigned int valid_channels,
  const arm_gemm::Requantize32 &qp
)
{
  struct Args
  {
    const int8_t *inptr;
    size_t ld_in_vl;
    long unsigned int pad_top, pad_bottom, pad_left;
    const int8_t *weights;
    long unsigned int input_cols, output_cols;
    int8_t **outptrs;
    const size_t *ld_out_cols;
    const size_t *ld_out_vls;
    long unsigned int current_channel, n_channels;
  };

  Args args = { inptr, ld_in_vl, pad_top, 8u - std::min(8u, pad_top + valid_input_rows), pad_left, weights, valid_input_cols, output_cols, outptrs, outlds, outvllds, start_channel, valid_channels };

  __asm__ __volatile__(
    ".inst 0xd503477f  // SMSTART ZA\n"
    "ldr x5, [%x[args], %[offsetof_Args_pad_bottom]]\n"
    "ptrue p2.b\n"
    "mov x19, #0x8\n"
    "ldr x6, [%x[args], %[offsetof_Args_pad_top]]\n"
    "ld1rh { z25.h }, p2/Z, [%x[qp], %[offsetof_Requantize32_a_offset]]\n"
    "sub x19, x19, x5\n"
    ".inst 0x25207812  // ptrue pn10.b\n"
    "ldr x7, [%x[args], %[offsetof_Args_n_channels]]\n"
    "whilelt p1.s, XZR, x7\n"
    "whilelt p9.s, XZR, x19\n"
    "ld1rw { z9.s }, p2/Z, [%x[qp], %[offsetof_Requantize32_c_offset]]\n"
    "whilelt p8.s, XZR, x6\n"
    "addvl SP, SP, #-30\n"
    "ldr x17, [%x[args], %[offsetof_Args_current_channel]]\n"
    "neg z25.h, p2/M, z25.h\n"
    "eor p8.b, p2/Z, p8.b, p9.b\n"
    "ld1rw { z3.s }, p2/Z, [%x[qp], %[offsetof_Requantize32_per_layer_mul]]\n"
    "ld1rw { z1.s }, p2/Z, [%x[qp], %[offsetof_Requantize32_per_layer_right_shift]]\n"
    "ld1rw { z24.s }, p2/Z, [%x[qp], %[offsetof_Requantize32_minval]]\n"
    "ld1rw { z31.s }, p2/Z, [%x[qp], %[offsetof_Requantize32_maxval]]\n"
    "1:"  // Channel loop
    "ldr x19, [%x[qp], %[offsetof_Requantize32_bias]]\n"
    "mov z6.s, #0x0\n"
    "cbz x19, 2f\n"
    "ld1w { z6.s }, p1/Z, [x19, x17, LSL #2]\n"
    "2:"  // Load bias: Done
    "ldr x23, [%x[args], %[offsetof_Args_weights]]\n"
    "mov x21, x23\n"
    "ld1sb { z18.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "ld1rh { z12.h }, p2/Z, [%x[qp], %[offsetof_Requantize32_b_offset]]\n"
    "mov z2.h, #0x0\n"
    "sub z18.h, z18.h, z12.h\n"
    "incw x23\n"
    "ld1sb { z17.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "sub z17.h, z17.h, z12.h\n"
    "trn1 z0.h, z2.h, z18.h\n"
    "ld1sb { z21.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "sub z21.h, z21.h, z12.h\n"
    "trn1 z8.h, z18.h, z17.h\n"
    "ld1sb { z16.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "sub z16.h, z16.h, z12.h\n"
    "trn1 z4.h, z17.h, z21.h\n"
    "ld1sb { z15.s }, p2/Z, [x21]\n"
    "sub z15.h, z15.h, z12.h\n"
    "mov x21, x23\n"
    "trn1 z5.h, z21.h, z16.h\n"
    "ld1sb { z18.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "trn1 z10.h, z16.h, z15.h\n"
    "trn1 z11.h, z15.h, z2.h\n"
    "ld1sb { z17.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "sub z18.h, z18.h, z12.h\n"
    "sub z17.h, z17.h, z12.h\n"
    "ld1sb { z21.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "sub z21.h, z21.h, z12.h\n"
    "addvl x20, SP, #30\n"
    "ld1sb { z16.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "incw x23\n"
    "sub z16.h, z16.h, z12.h\n"
    "ld1sb { z15.s }, p2/Z, [x21]\n"
    "addvl x20, x20, #-6\n"
    "sub z15.h, z15.h, z12.h\n"
    "mov x21, x23\n"
    "st1h { z0.h }, p2, [x20]\n"
    "trn1 z0.h, z2.h, z18.h\n"
    "incw x23\n"
    "ldr x19, [%x[qp], %[offsetof_Requantize32_per_channel_muls]]\n"
    "st1h { z8.h }, p2, [x20, #1, MUL VL]\n"
    "trn1 z8.h, z18.h, z17.h\n"
    "ld1sb { z18.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "st1h { z4.h }, p2, [x20, #2, MUL VL]\n"
    "trn1 z4.h, z17.h, z21.h\n"
    "ld1sb { z17.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "st1h { z5.h }, p2, [x20, #3, MUL VL]\n"
    "trn1 z5.h, z21.h, z16.h\n"
    "ld1sb { z21.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "st1h { z10.h }, p2, [x20, #4, MUL VL]\n"
    "trn1 z10.h, z16.h, z15.h\n"
    "ld1sb { z16.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "st1h { z11.h }, p2, [x20, #5, MUL VL]\n"
    "trn1 z11.h, z15.h, z2.h\n"
    "sub z18.h, z18.h, z12.h\n"
    "addvl x20, x20, #-6\n"
    "sub z17.h, z17.h, z12.h\n"
    "ld1sb { z15.s }, p2/Z, [x21]\n"
    "sub z21.h, z21.h, z12.h\n"
    "mov x21, x23\n"
    "sub z16.h, z16.h, z12.h\n"
    "sub z15.h, z15.h, z12.h\n"
    "st1h { z0.h }, p2, [x20]\n"
    "incw x23\n"
    "st1h { z8.h }, p2, [x20, #1, MUL VL]\n"
    "trn1 z0.h, z2.h, z18.h\n"
    "trn1 z8.h, z18.h, z17.h\n"
    "ld1sb { z18.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "st1h { z4.h }, p2, [x20, #2, MUL VL]\n"
    "trn1 z4.h, z17.h, z21.h\n"
    "ld1sb { z17.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "st1h { z5.h }, p2, [x20, #3, MUL VL]\n"
    "trn1 z5.h, z21.h, z16.h\n"
    "ld1sb { z21.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "st1h { z10.h }, p2, [x20, #4, MUL VL]\n"
    "trn1 z10.h, z16.h, z15.h\n"
    "ld1sb { z16.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "st1h { z11.h }, p2, [x20, #5, MUL VL]\n"
    "trn1 z11.h, z15.h, z2.h\n"
    "sub z18.h, z18.h, z12.h\n"
    "sub z17.h, z17.h, z12.h\n"
    "ld1sb { z15.s }, p2/Z, [x21]\n"
    "addvl x20, x20, #-6\n"
    "sub z21.h, z21.h, z12.h\n"
    "sub z16.h, z16.h, z12.h\n"
    "mov x21, x23\n"
    "st1h { z0.h }, p2, [x20]\n"
    "sub z15.h, z15.h, z12.h\n"
    "st1h { z8.h }, p2, [x20, #1, MUL VL]\n"
    "trn1 z0.h, z2.h, z18.h\n"
    "trn1 z8.h, z18.h, z17.h\n"
    "ld1sb { z18.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "st1h { z4.h }, p2, [x20, #2, MUL VL]\n"
    "trn1 z4.h, z17.h, z21.h\n"
    "ld1sb { z17.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "st1h { z5.h }, p2, [x20, #3, MUL VL]\n"
    "trn1 z5.h, z21.h, z16.h\n"
    "ld1sb { z21.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "st1h { z10.h }, p2, [x20, #4, MUL VL]\n"
    "trn1 z10.h, z16.h, z15.h\n"
    "ld1sb { z16.s }, p2/Z, [x21]\n"
    "incw x21, ALL, MUL #5\n"
    "st1h { z11.h }, p2, [x20, #5, MUL VL]\n"
    "trn1 z11.h, z15.h, z2.h\n"
    "ld1sb { z15.s }, p2/Z, [x21]\n"
    "sub z18.h, z18.h, z12.h\n"
    "addvl x20, x20, #-6\n"
    "sub z17.h, z17.h, z12.h\n"
    "sub z21.h, z21.h, z12.h\n"
    "st1h { z0.h }, p2, [x20]\n"
    "sub z16.h, z16.h, z12.h\n"
    "sub z15.h, z15.h, z12.h\n"
    "st1h { z8.h }, p2, [x20, #1, MUL VL]\n"
    "st1h { z4.h }, p2, [x20, #2, MUL VL]\n"
    "mov z7.d, z6.d\n"
    "trn1 z0.h, z2.h, z18.h\n"
    "st1h { z5.h }, p2, [x20, #3, MUL VL]\n"
    "trn1 z8.h, z18.h, z17.h\n"
    "trn1 z4.h, z17.h, z21.h\n"
    "st1h { z10.h }, p2, [x20, #4, MUL VL]\n"
    "trn1 z5.h, z21.h, z16.h\n"
    "trn1 z10.h, z16.h, z15.h\n"
    "st1h { z11.h }, p2, [x20, #5, MUL VL]\n"
    "addvl x20, x20, #-6\n"
    "trn1 z11.h, z15.h, z2.h\n"
    "st1h { z0.h }, p2, [x20]\n"
    "st1h { z8.h }, p2, [x20, #1, MUL VL]\n"
    "st1h { z4.h }, p2, [x20, #2, MUL VL]\n"
    "st1h { z5.h }, p2, [x20, #3, MUL VL]\n"
    "st1h { z10.h }, p2, [x20, #4, MUL VL]\n"
    "st1h { z11.h }, p2, [x20, #5, MUL VL]\n"
    "cbz x19, 3f\n"
    "ld1w { z3.s }, p1/Z, [x19, x17, LSL #2]\n"
    "3:"  // Load mul: End
    "ldr x19, [%x[qp], %[offsetof_Requantize32_per_channel_right_shifts]]\n"
    "cbz x19, 4f\n"
    "ld1w { z1.s }, p1/Z, [x19, x17, LSL #2]\n"
    "4:"  // Load right_shift: End
    "ldr x16, [%x[args], %[offsetof_Args_input_cols]]\n"
    "sub x19, x16, #0x1\n"
    "orr x22, x19, %x[ld_in_col], LSL #16\n"
    "ldr x15, [%x[args], %[offsetof_Args_inptr]]\n"
    "orr x22, x7, x22, LSL #22\n"
    "mov x21, #0x8\n"
    "add x20, x6, x5\n"
    "lsl x19, %x[ld_in_row], #0x0\n"
    "ldr x14, [%x[args], %[offsetof_Args_output_cols]]\n"
    "mov x11, #0x0\n"
    "mov x8, #0x8\n"
    "lsl x22, x22, #0x0\n"
    "sub x21, x21, x20\n"
    "madd x19, x19, x6, x15\n"
    "5:"  // Issue prefetches
    "subs x21, x21, #0x1\n"
    ".inst 0xf8b64a7c  // rprfm pldstrm, x22, [x19]\n"
    "add x19, x19, %x[ld_in_col]\n"
    "bgt 5b\n"
    "ldr x24, [%x[args], %[offsetof_Args_outptrs]]\n"
    "lsl x19, %x[ld_in_row], #0x0\n"
    "msub x15, x6, x19, x15\n"
    ".inst 0xc00468c0  // mova za.d[x11, #0], { z6.d-z7.d }\n"
    "ldr x19, [%x[args], %[offsetof_Args_ld_out_cols]]\n"
    ".inst 0xc00468c1  // mova za.d[x11, #1], { z6.d-z7.d }\n"
    "mov x21, #0x4\n"
    "ldp x13, x4, [x24], #0x10\n"
    ".inst 0xc00468c2  // mova za.d[x11, #2], { z6.d-z7.d }\n"
    "ldp x10, x9, [x19], #0x10\n"
    ".inst 0xc00468c3  // mova za.d[x11, #3], { z6.d-z7.d }\n"
    "ldr x20, [%x[args], %[offsetof_Args_pad_left]]\n"
    ".inst 0xc00468c4  // mova za.d[x11, #4], { z6.d-z7.d }\n"
    "ldp x28, x27, [x24], #0x10\n"
    ".inst 0xc00468c5  // mova za.d[x11, #5], { z6.d-z7.d }\n"
    "ldp x26, x25, [x19], #0x10\n"
    ".inst 0xc00468c6  // mova za.d[x11, #6], { z6.d-z7.d }\n"
    ".inst 0xc00468c7  // mova za.d[x11, #7], { z6.d-z7.d }\n"
    ".inst 0xc00408c0  // mova za.d[x8, #0], { z6.d-z7.d }\n"
    ".inst 0xc00408c1  // mova za.d[x8, #1], { z6.d-z7.d }\n"
    "cbz x20, 7f\n"
    "cmp x20, x21\n"
    "csel x19, x20, x21, LT\n"
    "sub x20, x20, x19\n"
    "sub x21, x21, x19\n"
    "cbz x20, 7f\n"
    ".inst 0xc006680c  // mova { z12.d-z13.d }, za.d[x11, #0]\n"
    "sub x14, x14, x20\n"
    ".inst 0xc006682e  // mova { z14.d-z15.d }, za.d[x11, #1]\n"
    ".inst 0xc1a3ac0c  // sqdmulh { z12.s-z15.s }, { z12.s-z15.s }, z3.s\n"
    ".inst 0xc1a1aa2c  // srshl { z12.s-z15.s }, { z12.s-z15.s }, z1.s\n"
    ".inst 0xc1a9ab0c  // add { z12.s-z15.s }, { z12.s-z15.s }, z9.s\n"
    ".inst 0xc1bfcf0c  // sclamp { z12.s-z15.s }, z24.s, z31.s\n"
    "6:"  // Left padding
    "subs x20, x20, #0x1\n"
    "st1b { z12.s }, p1, [x13]\n"
    "add x13, x13, x10\n"
    "st1b { z14.s }, p1, [x4]\n"
    "add x4, x4, x9\n"
    "st1b { z13.s }, p1, [x28]\n"
    "add x28, x28, x26\n"
    "st1b { z15.s }, p1, [x27]\n"
    "add x27, x27, x25\n"
    "bgt 6b\n"
    "7:"  // Left padding: End
    "adds XZR, x6, x5\n"
    "bne 14f\n"
    "cbz x21, 12f\n"
    "cmp x21, #0x1\n"
    "sub x16, x16, x21\n"
    "beq 11f\n"
    "cmp x21, #0x2\n"
    "beq 10f\n"
    "cmp x21, #0x3\n"
    "beq 9f\n"
    "8:"  // Unpadded: 4 priming loads
    "add x20, x15, %x[ld_in_row]\n"
    "ld1sb { z17.s }, p1/Z, [x15]\n"
    "addvl x19, SP, #24\n"
    "ld1sb { z16.s }, p1/Z, [x20]\n"
    "add x20, x20, %x[ld_in_row]\n"
    "trn1 z27.h, z17.h, z16.h\n"
    "add z27.h, z27.h, z25.h\n"
    "ld1sb { z17.s }, p1/Z, [x20]\n"
    "add x20, x20, %x[ld_in_row]\n"
    "add x15, x15, %x[ld_in_col]\n"
    "ld1sb { z16.s }, p1/Z, [x20]\n"
    "add x20, x20, %x[ld_in_row]\n"
    "trn1 z28.h, z17.h, z16.h\n"
    "add z28.h, z28.h, z25.h\n"
    "ld1sb { z16.s }, p1/Z, [x20]\n"
    "add x20, x20, %x[ld_in_row]\n"
    "ld1sb { z29.s }, p1/Z, [x20]\n"
    "add x20, x20, %x[ld_in_row]\n"
    "trn1 z29.h, z16.h, z29.h\n"
    "add z29.h, z29.h, z25.h\n"
    "ld1sb { z17.s }, p1/Z, [x20]\n"
    "add x20, x20, %x[ld_in_row]\n"
    ".inst 0xa1402a60  // ld1h { z0.h, z8.h }, pn10.b/Z, [x19]\n"
    ".inst 0xc1687768  // sdot za.s[x11, 0], { z27.h-z28.h }, z8.h\n"
    "ld1sb { z16.s }, p1/Z, [x20]\n"
    "trn1 z30.h, z17.h, z16.h\n"
    ".inst 0xc1607769  // sdot za.s[x11, 1], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa0412a64  // ld1h { z4.h-z5.h }, pn10.b/Z, [x19, #0x2, MUL VL]\n"
    "add z30.h, z30.h, z25.h\n"
    ".inst 0xc1657788  // sdot za.s[x11, 0], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc1647789  // sdot za.s[x11, 1], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0422a6a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x19, #0x4, MUL VL]\n"
    ".inst 0xc16b77a8  // sdot za.s[x11, 0], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77a9  // sdot za.s[x11, 1], { z29.h-z30.h }, z10.h\n"
    "9:"  // Unpadded: 3 priming loads
    "add x21, x15, %x[ld_in_row]\n"
    "ld1sb { z17.s }, p1/Z, [x15]\n"
    "addvl x20, SP, #18\n"
    "ld1sb { z16.s }, p1/Z, [x21]\n"
    "add x21, x21, %x[ld_in_row]\n"
    "trn1 z27.h, z17.h, z16.h\n"
    "add z27.h, z27.h, z25.h\n"
    "ld1sb { z17.s }, p1/Z, [x21]\n"
    "add x21, x21, %x[ld_in_row]\n"
    "addvl x19, SP, #24\n"
    "ld1sb { z16.s }, p1/Z, [x21]\n"
    "add x21, x21, %x[ld_in_row]\n"
    "trn1 z28.h, z17.h, z16.h\n"
    "add z28.h, z28.h, z25.h\n"
    "ld1sb { z17.s }, p1/Z, [x21]\n"
    "add x21, x21, %x[ld_in_row]\n"
    "add x15, x15, %x[ld_in_col]\n"
    "ld1sb { z16.s }, p1/Z, [x21]\n"
    "add x21, x21, %x[ld_in_row]\n"
    "trn1 z29.h, z17.h, z16.h\n"
    "add z29.h, z29.h, z25.h\n"
    "ld1sb { z17.s }, p1/Z, [x21]\n"
    "add x21, x21, %x[ld_in_row]\n"
    ".inst 0xa1402a80  // ld1h { z0.h, z8.h }, pn10.b/Z, [x20]\n"
    ".inst 0xc1687768  // sdot za.s[x11, 0], { z27.h-z28.h }, z8.h\n"
    "ld1sb { z16.s }, p1/Z, [x21]\n"
    "trn1 z30.h, z17.h, z16.h\n"
    ".inst 0xc1607769  // sdot za.s[x11, 1], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402a60  // ld1h { z0.h, z8.h }, pn10.b/Z, [x19]\n"
    ".inst 0xc168776a  // sdot za.s[x11, 2], { z27.h-z28.h }, z8.h\n"
    "add z30.h, z30.h, z25.h\n"
    ".inst 0xa0412a84  // ld1h { z4.h-z5.h }, pn10.b/Z, [x20, #0x2, MUL VL]\n"
    ".inst 0xc160776b  // sdot za.s[x11, 3], { z27.h-z28.h }, z0.h\n"
    ".inst 0xc1657788  // sdot za.s[x11, 0], { z28.h-z29.h }, z5.h\n"
    ".inst 0xa0422a8a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x20, #0x4, MUL VL]\n"
    ".inst 0xc1647789  // sdot za.s[x11, 1], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412a64  // ld1h { z4.h-z5.h }, pn10.b/Z, [x19, #0x2, MUL VL]\n"
    ".inst 0xc165778a  // sdot za.s[x11, 2], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778b  // sdot za.s[x11, 3], { z28.h-z29.h }, z4.h\n"
    ".inst 0xc16b77a8  // sdot za.s[x11, 0], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77a9  // sdot za.s[x11, 1], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422a6a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x19, #0x4, MUL VL]\n"
    ".inst 0xc16b77aa  // sdot za.s[x11, 2], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ab  // sdot za.s[x11, 3], { z29.h-z30.h }, z10.h\n"
    "10:"  // Unpadded: 2 priming loads
    "add x22, x15, %x[ld_in_row]\n"
    "ld1sb { z17.s }, p1/Z, [x15]\n"
    "addvl x21, SP, #12\n"
    "ld1sb { z16.s }, p1/Z, [x22]\n"
    "add x22, x22, %x[ld_in_row]\n"
    "trn1 z27.h, z17.h, z16.h\n"
    "add z27.h, z27.h, z25.h\n"
    "ld1sb { z17.s }, p1/Z, [x22]\n"
    "add x22, x22, %x[ld_in_row]\n"
    "addvl x20, SP, #18\n"
    "ld1sb { z16.s }, p1/Z, [x22]\n"
    "add x22, x22, %x[ld_in_row]\n"
    "trn1 z28.h, z17.h, z16.h\n"
    "add z28.h, z28.h, z25.h\n"
    "ld1sb { z17.s }, p1/Z, [x22]\n"
    "add x22, x22, %x[ld_in_row]\n"
    "addvl x19, SP, #24\n"
    "ld1sb { z16.s }, p1/Z, [x22]\n"
    "add x22, x22, %x[ld_in_row]\n"
    "trn1 z29.h, z17.h, z16.h\n"
    "add z29.h, z29.h, z25.h\n"
    "ld1sb { z17.s }, p1/Z, [x22]\n"
    "add x22, x22, %x[ld_in_row]\n"
    "add x15, x15, %x[ld_in_col]\n"
    ".inst 0xa1402aa0  // ld1h { z0.h, z8.h }, pn10.b/Z, [x21]\n"
    ".inst 0xc1687768  // sdot za.s[x11, 0], { z27.h-z28.h }, z8.h\n"
    "ld1sb { z16.s }, p1/Z, [x22]\n"
    "trn1 z30.h, z17.h, z16.h\n"
    ".inst 0xc1607769  // sdot za.s[x11, 1], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402a80  // ld1h { z0.h, z8.h }, pn10.b/Z, [x20]\n"
    ".inst 0xc168776a  // sdot za.s[x11, 2], { z27.h-z28.h }, z8.h\n"
    "add z30.h, z30.h, z25.h\n"
    ".inst 0xa0412aa4  // ld1h { z4.h-z5.h }, pn10.b/Z, [x21, #0x2, MUL VL]\n"
    ".inst 0xc160776b  // sdot za.s[x11, 3], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402a60  // ld1h { z0.h, z8.h }, pn10.b/Z, [x19]\n"
    ".inst 0xc1657788  // sdot za.s[x11, 0], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc1647789  // sdot za.s[x11, 1], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412a84  // ld1h { z4.h-z5.h }, pn10.b/Z, [x20, #0x2, MUL VL]\n"
    ".inst 0xa0422aaa  // ld1h { z10.h-z11.h }, pn10.b/Z, [x21, #0x4, MUL VL]\n"
    ".inst 0xc168776c  // sdot za.s[x11, 4], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc160776d  // sdot za.s[x11, 5], { z27.h-z28.h }, z0.h\n"
    ".inst 0xc165778a  // sdot za.s[x11, 2], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778b  // sdot za.s[x11, 3], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412a64  // ld1h { z4.h-z5.h }, pn10.b/Z, [x19, #0x2, MUL VL]\n"
    ".inst 0xc16b77a8  // sdot za.s[x11, 0], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77a9  // sdot za.s[x11, 1], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422a8a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x20, #0x4, MUL VL]\n"
    ".inst 0xc165778c  // sdot za.s[x11, 4], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778d  // sdot za.s[x11, 5], { z28.h-z29.h }, z4.h\n"
    ".inst 0xc16b77aa  // sdot za.s[x11, 2], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ab  // sdot za.s[x11, 3], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422a6a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x19, #0x4, MUL VL]\n"
    ".inst 0xc16b77ac  // sdot za.s[x11, 4], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ad  // sdot za.s[x11, 5], { z29.h-z30.h }, z10.h\n"
    "11:"  // Unpadded: 1 priming loads
    "add x23, x15, %x[ld_in_row]\n"
    "ld1sb { z17.s }, p1/Z, [x15]\n"
    "addvl x22, SP, #6\n"
    "ld1sb { z16.s }, p1/Z, [x23]\n"
    "add x23, x23, %x[ld_in_row]\n"
    "trn1 z27.h, z17.h, z16.h\n"
    "add z27.h, z27.h, z25.h\n"
    "ld1sb { z17.s }, p1/Z, [x23]\n"
    "add x23, x23, %x[ld_in_row]\n"
    "addvl x21, SP, #12\n"
    "ld1sb { z16.s }, p1/Z, [x23]\n"
    "add x23, x23, %x[ld_in_row]\n"
    "trn1 z28.h, z17.h, z16.h\n"
    "add z28.h, z28.h, z25.h\n"
    "ld1sb { z17.s }, p1/Z, [x23]\n"
    "add x23, x23, %x[ld_in_row]\n"
    "addvl x20, SP, #18\n"
    "ld1sb { z16.s }, p1/Z, [x23]\n"
    "add x23, x23, %x[ld_in_row]\n"
    "trn1 z29.h, z17.h, z16.h\n"
    "add z29.h, z29.h, z25.h\n"
    "ld1sb { z17.s }, p1/Z, [x23]\n"
    "add x23, x23, %x[ld_in_row]\n"
    "addvl x19, SP, #24\n"
    ".inst 0xa1402ac0  // ld1h { z0.h, z8.h }, pn10.b/Z, [x22]\n"
    ".inst 0xc1687768  // sdot za.s[x11, 0], { z27.h-z28.h }, z8.h\n"
    "add x15, x15, %x[ld_in_col]\n"
    "ld1sb { z16.s }, p1/Z, [x23]\n"
    "trn1 z30.h, z17.h, z16.h\n"
    ".inst 0xc1607769  // sdot za.s[x11, 1], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402aa0  // ld1h { z0.h, z8.h }, pn10.b/Z, [x21]\n"
    ".inst 0xc168776a  // sdot za.s[x11, 2], { z27.h-z28.h }, z8.h\n"
    "add z30.h, z30.h, z25.h\n"
    ".inst 0xa0412ac4  // ld1h { z4.h-z5.h }, pn10.b/Z, [x22, #0x2, MUL VL]\n"
    ".inst 0xc160776b  // sdot za.s[x11, 3], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402a80  // ld1h { z0.h, z8.h }, pn10.b/Z, [x20]\n"
    ".inst 0xc1657788  // sdot za.s[x11, 0], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc1647789  // sdot za.s[x11, 1], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412aa4  // ld1h { z4.h-z5.h }, pn10.b/Z, [x21, #0x2, MUL VL]\n"
    ".inst 0xa0422aca  // ld1h { z10.h-z11.h }, pn10.b/Z, [x22, #0x4, MUL VL]\n"
    ".inst 0xc168776c  // sdot za.s[x11, 4], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc160776d  // sdot za.s[x11, 5], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402a60  // ld1h { z0.h, z8.h }, pn10.b/Z, [x19]\n"
    ".inst 0xc165778a  // sdot za.s[x11, 2], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778b  // sdot za.s[x11, 3], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412a84  // ld1h { z4.h-z5.h }, pn10.b/Z, [x20, #0x2, MUL VL]\n"
    ".inst 0xc16b77a8  // sdot za.s[x11, 0], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77a9  // sdot za.s[x11, 1], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422aaa  // ld1h { z10.h-z11.h }, pn10.b/Z, [x21, #0x4, MUL VL]\n"
    ".inst 0xc168776e  // sdot za.s[x11, 6], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc160776f  // sdot za.s[x11, 7], { z27.h-z28.h }, z0.h\n"
    ".inst 0xc165778c  // sdot za.s[x11, 4], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778d  // sdot za.s[x11, 5], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412a64  // ld1h { z4.h-z5.h }, pn10.b/Z, [x19, #0x2, MUL VL]\n"
    ".inst 0xc16b77aa  // sdot za.s[x11, 2], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ab  // sdot za.s[x11, 3], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422a8a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x20, #0x4, MUL VL]\n"
    ".inst 0xc165778e  // sdot za.s[x11, 6], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778f  // sdot za.s[x11, 7], { z28.h-z29.h }, z4.h\n"
    ".inst 0xc16b77ac  // sdot za.s[x11, 4], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ad  // sdot za.s[x11, 5], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422a6a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x19, #0x4, MUL VL]\n"
    ".inst 0xc16b77ae  // sdot za.s[x11, 6], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77af  // sdot za.s[x11, 7], { z29.h-z30.h }, z10.h\n"
    "12:"  // Unpadded: 0 priming loads
    ".inst 0xa1402be0  // ld1h { z0.h, z8.h }, pn10.b/Z, [SP]\n"
    ".inst 0xa0412be4  // ld1h { z4.h-z5.h }, pn10.b/Z, [SP, #0x2, MUL VL]\n"
    ".inst 0xa0422bea  // ld1h { z10.h-z11.h }, pn10.b/Z, [SP, #0x4, MUL VL]\n"
    "cbz x16, 22f\n"
    "add x19, x15, %x[ld_in_row]\n"
    "ld1sb { z17.s }, p1/Z, [x15]\n"
    "sub x16, x16, #0x1\n"
    "ld1sb { z16.s }, p1/Z, [x19]\n"
    "add x19, x19, %x[ld_in_row]\n"
    "trn1 z27.h, z17.h, z16.h\n"
    "sub x14, x14, #0x1\n"
    "ld1sb { z17.s }, p1/Z, [x19]\n"
    "add x19, x19, %x[ld_in_row]\n"
    "cmp x16, x14\n"
    "add z27.h, z27.h, z25.h\n"
    "ld1sb { z16.s }, p1/Z, [x19]\n"
    "add x19, x19, %x[ld_in_row]\n"
    "trn1 z28.h, z17.h, z16.h\n"
    "csel x24, x16, x14, LT\n"
    "ld1sb { z17.s }, p1/Z, [x19]\n"
    "add x19, x19, %x[ld_in_row]\n"
    "add z28.h, z28.h, z25.h\n"
    "add x15, x15, %x[ld_in_col]\n"
    "ld1sb { z16.s }, p1/Z, [x19]\n"
    "add x19, x19, %x[ld_in_row]\n"
    "trn1 z29.h, z17.h, z16.h\n"
    "add z29.h, z29.h, z25.h\n"
    "ld1sb { z17.s }, p1/Z, [x19]\n"
    "add x19, x19, %x[ld_in_row]\n"
    "sub x14, x14, x24\n"
    "ld1sb { z16.s }, p1/Z, [x19]\n"
    "trn1 z30.h, z17.h, z16.h\n"
    "add z30.h, z30.h, z25.h\n"
    "cbz x24, 21f\n"
    "13:"  // Unpadded: Main loop
    "addvl x23, SP, #6\n"
    ".inst 0xc1687768  // sdot za.s[x11, 0], { z27.h-z28.h }, z8.h\n"
    "addvl x22, SP, #12\n"
    "ld1sb { z23.s }, p1/Z, [x15]\n"
    ".inst 0xc1607769  // sdot za.s[x11, 1], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402ae0  // ld1h { z0.h, z8.h }, pn10.b/Z, [x23]\n"
    "addvl x21, SP, #18\n"
    "addvl x20, SP, #24\n"
    ".inst 0xc168776a  // sdot za.s[x11, 2], { z27.h-z28.h }, z8.h\n"
    "add x19, x15, %x[ld_in_row]\n"
    "ld1sb { z22.s }, p1/Z, [x19]\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0xc160776b  // sdot za.s[x11, 3], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402ac0  // ld1h { z0.h, z8.h }, pn10.b/Z, [x22]\n"
    "subs x24, x24, #0x1\n"
    "add x15, x15, %x[ld_in_col]\n"
    ".inst 0xc1657788  // sdot za.s[x11, 0], { z28.h-z29.h }, z5.h\n"
    "ld1sb { z21.s }, p1/Z, [x19]\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0xc1647789  // sdot za.s[x11, 1], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412ae4  // ld1h { z4.h-z5.h }, pn10.b/Z, [x23, #0x2, MUL VL]\n"
    ".inst 0xc168776c  // sdot za.s[x11, 4], { z27.h-z28.h }, z8.h\n"
    "ld1sb { z20.s }, p1/Z, [x19]\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0xc160776d  // sdot za.s[x11, 5], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402aa0  // ld1h { z0.h, z8.h }, pn10.b/Z, [x21]\n"
    ".inst 0xc165778a  // sdot za.s[x11, 2], { z28.h-z29.h }, z5.h\n"
    "ld1sb { z19.s }, p1/Z, [x19]\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0xc164778b  // sdot za.s[x11, 3], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412ac4  // ld1h { z4.h-z5.h }, pn10.b/Z, [x22, #0x2, MUL VL]\n"
    ".inst 0xc16b77a8  // sdot za.s[x11, 0], { z29.h-z30.h }, z11.h\n"
    "ld1sb { z18.s }, p1/Z, [x19]\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0xc16a77a9  // sdot za.s[x11, 1], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422aea  // ld1h { z10.h-z11.h }, pn10.b/Z, [x23, #0x4, MUL VL]\n"
    ".inst 0xc168776e  // sdot za.s[x11, 6], { z27.h-z28.h }, z8.h\n"
    "ld1sb { z17.s }, p1/Z, [x19]\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0xc160776f  // sdot za.s[x11, 7], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402a80  // ld1h { z0.h, z8.h }, pn10.b/Z, [x20]\n"
    ".inst 0xc165778c  // sdot za.s[x11, 4], { z28.h-z29.h }, z5.h\n"
    "ld1sb { z16.s }, p1/Z, [x19]\n"
    ".inst 0xc164778d  // sdot za.s[x11, 5], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412aa4  // ld1h { z4.h-z5.h }, pn10.b/Z, [x21, #0x2, MUL VL]\n"
    ".inst 0xc16b77aa  // sdot za.s[x11, 2], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ab  // sdot za.s[x11, 3], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422aca  // ld1h { z10.h-z11.h }, pn10.b/Z, [x22, #0x4, MUL VL]\n"
    ".inst 0xc165778e  // sdot za.s[x11, 6], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778f  // sdot za.s[x11, 7], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412a84  // ld1h { z4.h-z5.h }, pn10.b/Z, [x20, #0x2, MUL VL]\n"
    ".inst 0xc16b77ac  // sdot za.s[x11, 4], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ad  // sdot za.s[x11, 5], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422aaa  // ld1h { z10.h-z11.h }, pn10.b/Z, [x21, #0x4, MUL VL]\n"
    ".inst 0xc16b77ae  // sdot za.s[x11, 6], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77af  // sdot za.s[x11, 7], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422a8a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x20, #0x4, MUL VL]\n"
    ".inst 0xc1681768  // sdot za.s[x8, 0], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc1601769  // sdot za.s[x8, 1], { z27.h-z28.h }, z0.h\n"
    "trn1 z27.h, z23.h, z22.h\n"
    ".inst 0xa1402be0  // ld1h { z0.h, z8.h }, pn10.b/Z, [SP]\n"
    ".inst 0xc1651788  // sdot za.s[x8, 0], { z28.h-z29.h }, z5.h\n"
    "add z27.h, z27.h, z25.h\n"
    ".inst 0xc1641789  // sdot za.s[x8, 1], { z28.h-z29.h }, z4.h\n"
    "trn1 z28.h, z21.h, z20.h\n"
    ".inst 0xa0412be4  // ld1h { z4.h-z5.h }, pn10.b/Z, [SP, #0x2, MUL VL]\n"
    ".inst 0xc16b17a8  // sdot za.s[x8, 0], { z29.h-z30.h }, z11.h\n"
    "add z28.h, z28.h, z25.h\n"
    ".inst 0xc16a17a9  // sdot za.s[x8, 1], { z29.h-z30.h }, z10.h\n"
    "trn1 z29.h, z19.h, z18.h\n"
    "trn1 z30.h, z17.h, z16.h\n"
    "add x8, x8, #0x2\n"
    ".inst 0xc006680c  // mova { z12.d-z13.d }, za.d[x11, #0]\n"
    ".inst 0xa0422bea  // ld1h { z10.h-z11.h }, pn10.b/Z, [SP, #0x4, MUL VL]\n"
    "add z29.h, z29.h, z25.h\n"
    ".inst 0xc006682e  // mova { z14.d-z15.d }, za.d[x11, #1]\n"
    ".inst 0xc1a3ac0c  // sqdmulh { z12.s-z15.s }, { z12.s-z15.s }, z3.s\n"
    "add x11, x11, #0x2\n"
    ".inst 0xc1a1aa2c  // srshl { z12.s-z15.s }, { z12.s-z15.s }, z1.s\n"
    ".inst 0xc00408c0  // mova za.d[x8, #0], { z6.d-z7.d }\n"
    ".inst 0xc1a9ab0c  // add { z12.s-z15.s }, { z12.s-z15.s }, z9.s\n"
    ".inst 0xc00408c1  // mova za.d[x8, #1], { z6.d-z7.d }\n"
    ".inst 0xc1bfcf0c  // sclamp { z12.s-z15.s }, z24.s, z31.s\n"
    "st1b { z12.s }, p1, [x13]\n"
    "add x13, x13, x10\n"
    "add z30.h, z30.h, z25.h\n"
    "st1b { z14.s }, p1, [x4]\n"
    "add x4, x4, x9\n"
    "st1b { z13.s }, p1, [x28]\n"
    "add x28, x28, x26\n"
    "st1b { z15.s }, p1, [x27]\n"
    "add x27, x27, x25\n"
    "bgt 13b\n"
    "b 21f\n"
    "14:"  // Padded
    "cbz x21, 19f\n"
    "cmp x21, #0x1\n"
    "sub x16, x16, x21\n"
    "beq 18f\n"
    "cmp x21, #0x2\n"
    "beq 17f\n"
    "cmp x21, #0x3\n"
    "beq 16f\n"
    "15:"  // Padded: 4 priming loads
    "mov x12, #0x0\n"
    ".inst 0x25304500  // psel p0.s, p1.s/Z, p8.s[w12]\n"
    "ld1sb { z19.s }, p0/Z, [x15]\n"
    "add z19.h, p0/M, z19.h, z25.h\n"
    "add x20, x15, %x[ld_in_row]\n"
    ".inst 0x25704500  // psel p0.s, p1.s/Z, p8.s[w12, #1]\n"
    "ld1sb { z18.s }, p0/Z, [x20]\n"
    "add z18.h, p0/M, z18.h, z25.h\n"
    "add x20, x20, %x[ld_in_row]\n"
    ".inst 0x25b04500  // psel p0.s, p1.s/Z, p8.s[w12, #2]\n"
    "ld1sb { z17.s }, p0/Z, [x20]\n"
    "add z17.h, p0/M, z17.h, z25.h\n"
    "add x20, x20, %x[ld_in_row]\n"
    ".inst 0x25f04500  // psel p0.s, p1.s/Z, p8.s[w12, #3]\n"
    "ld1sb { z16.s }, p0/Z, [x20]\n"
    "add z16.h, p0/M, z16.h, z25.h\n"
    "mov x12, #0x4\n"
    "add x20, x20, %x[ld_in_row]\n"
    "trn1 z27.h, z19.h, z18.h\n"
    "trn1 z28.h, z17.h, z16.h\n"
    ".inst 0x25304500  // psel p0.s, p1.s/Z, p8.s[w12]\n"
    "ld1sb { z18.s }, p0/Z, [x20]\n"
    "add z18.h, p0/M, z18.h, z25.h\n"
    "add x20, x20, %x[ld_in_row]\n"
    ".inst 0x25704500  // psel p0.s, p1.s/Z, p8.s[w12, #1]\n"
    "ld1sb { z16.s }, p0/Z, [x20]\n"
    "add z16.h, p0/M, z16.h, z25.h\n"
    "add x20, x20, %x[ld_in_row]\n"
    ".inst 0x25b04500  // psel p0.s, p1.s/Z, p8.s[w12, #2]\n"
    "ld1sb { z17.s }, p0/Z, [x20]\n"
    "addvl x19, SP, #24\n"
    "add z17.h, p0/M, z17.h, z25.h\n"
    "add x20, x20, %x[ld_in_row]\n"
    ".inst 0x25f04500  // psel p0.s, p1.s/Z, p8.s[w12, #3]\n"
    ".inst 0xa1402a60  // ld1h { z0.h, z8.h }, pn10.b/Z, [x19]\n"
    "trn1 z29.h, z18.h, z16.h\n"
    "ld1sb { z16.s }, p0/Z, [x20]\n"
    "add z16.h, p0/M, z16.h, z25.h\n"
    ".inst 0xc1687768  // sdot za.s[x11, 0], { z27.h-z28.h }, z8.h\n"
    "add x15, x15, %x[ld_in_col]\n"
    ".inst 0xc1607769  // sdot za.s[x11, 1], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa0412a64  // ld1h { z4.h-z5.h }, pn10.b/Z, [x19, #0x2, MUL VL]\n"
    "trn1 z30.h, z17.h, z16.h\n"
    ".inst 0xc1657788  // sdot za.s[x11, 0], { z28.h-z29.h }, z5.h\n"
    ".inst 0xa0422a6a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x19, #0x4, MUL VL]\n"
    ".inst 0xc1647789  // sdot za.s[x11, 1], { z28.h-z29.h }, z4.h\n"
    ".inst 0xc16b77a8  // sdot za.s[x11, 0], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77a9  // sdot za.s[x11, 1], { z29.h-z30.h }, z10.h\n"
    "16:"  // Padded: 3 priming loads
    "mov x12, #0x0\n"
    ".inst 0x25304500  // psel p0.s, p1.s/Z, p8.s[w12]\n"
    "ld1sb { z19.s }, p0/Z, [x15]\n"
    "add z19.h, p0/M, z19.h, z25.h\n"
    "add x19, x15, %x[ld_in_row]\n"
    ".inst 0x25704500  // psel p0.s, p1.s/Z, p8.s[w12, #1]\n"
    "ld1sb { z18.s }, p0/Z, [x19]\n"
    "add z18.h, p0/M, z18.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25b04500  // psel p0.s, p1.s/Z, p8.s[w12, #2]\n"
    "ld1sb { z17.s }, p0/Z, [x19]\n"
    "add z17.h, p0/M, z17.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25f04500  // psel p0.s, p1.s/Z, p8.s[w12, #3]\n"
    "ld1sb { z16.s }, p0/Z, [x19]\n"
    "add z16.h, p0/M, z16.h, z25.h\n"
    "mov x12, #0x4\n"
    "add x19, x19, %x[ld_in_row]\n"
    "trn1 z27.h, z19.h, z18.h\n"
    "trn1 z28.h, z17.h, z16.h\n"
    ".inst 0x25304500  // psel p0.s, p1.s/Z, p8.s[w12]\n"
    "ld1sb { z18.s }, p0/Z, [x19]\n"
    "add z18.h, p0/M, z18.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25704500  // psel p0.s, p1.s/Z, p8.s[w12, #1]\n"
    "ld1sb { z16.s }, p0/Z, [x19]\n"
    "add z16.h, p0/M, z16.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25b04500  // psel p0.s, p1.s/Z, p8.s[w12, #2]\n"
    "ld1sb { z17.s }, p0/Z, [x19]\n"
    "addvl x20, SP, #18\n"
    "add z17.h, p0/M, z17.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25f04500  // psel p0.s, p1.s/Z, p8.s[w12, #3]\n"
    ".inst 0xa1402a80  // ld1h { z0.h, z8.h }, pn10.b/Z, [x20]\n"
    "trn1 z29.h, z18.h, z16.h\n"
    "ld1sb { z16.s }, p0/Z, [x19]\n"
    "addvl x19, SP, #24\n"
    "add z16.h, p0/M, z16.h, z25.h\n"
    ".inst 0xc1687768  // sdot za.s[x11, 0], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc1607769  // sdot za.s[x11, 1], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402a60  // ld1h { z0.h, z8.h }, pn10.b/Z, [x19]\n"
    "trn1 z30.h, z17.h, z16.h\n"
    "add x15, x15, %x[ld_in_col]\n"
    ".inst 0xa0412a84  // ld1h { z4.h-z5.h }, pn10.b/Z, [x20, #0x2, MUL VL]\n"
    ".inst 0xc168776a  // sdot za.s[x11, 2], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc160776b  // sdot za.s[x11, 3], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa0422a8a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x20, #0x4, MUL VL]\n"
    ".inst 0xc1657788  // sdot za.s[x11, 0], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc1647789  // sdot za.s[x11, 1], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412a64  // ld1h { z4.h-z5.h }, pn10.b/Z, [x19, #0x2, MUL VL]\n"
    ".inst 0xc165778a  // sdot za.s[x11, 2], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778b  // sdot za.s[x11, 3], { z28.h-z29.h }, z4.h\n"
    ".inst 0xc16b77a8  // sdot za.s[x11, 0], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77a9  // sdot za.s[x11, 1], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422a6a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x19, #0x4, MUL VL]\n"
    ".inst 0xc16b77aa  // sdot za.s[x11, 2], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ab  // sdot za.s[x11, 3], { z29.h-z30.h }, z10.h\n"
    "17:"  // Padded: 2 priming loads
    "mov x12, #0x0\n"
    ".inst 0x25304500  // psel p0.s, p1.s/Z, p8.s[w12]\n"
    "ld1sb { z19.s }, p0/Z, [x15]\n"
    "add z19.h, p0/M, z19.h, z25.h\n"
    "add x19, x15, %x[ld_in_row]\n"
    ".inst 0x25704500  // psel p0.s, p1.s/Z, p8.s[w12, #1]\n"
    "ld1sb { z18.s }, p0/Z, [x19]\n"
    "add z18.h, p0/M, z18.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25b04500  // psel p0.s, p1.s/Z, p8.s[w12, #2]\n"
    "ld1sb { z17.s }, p0/Z, [x19]\n"
    "add z17.h, p0/M, z17.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25f04500  // psel p0.s, p1.s/Z, p8.s[w12, #3]\n"
    "ld1sb { z16.s }, p0/Z, [x19]\n"
    "add z16.h, p0/M, z16.h, z25.h\n"
    "mov x12, #0x4\n"
    "add x19, x19, %x[ld_in_row]\n"
    "trn1 z27.h, z19.h, z18.h\n"
    "trn1 z28.h, z17.h, z16.h\n"
    ".inst 0x25304500  // psel p0.s, p1.s/Z, p8.s[w12]\n"
    "ld1sb { z18.s }, p0/Z, [x19]\n"
    "add z18.h, p0/M, z18.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25704500  // psel p0.s, p1.s/Z, p8.s[w12, #1]\n"
    "ld1sb { z16.s }, p0/Z, [x19]\n"
    "add z16.h, p0/M, z16.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25b04500  // psel p0.s, p1.s/Z, p8.s[w12, #2]\n"
    "ld1sb { z17.s }, p0/Z, [x19]\n"
    "addvl x21, SP, #12\n"
    "add z17.h, p0/M, z17.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25f04500  // psel p0.s, p1.s/Z, p8.s[w12, #3]\n"
    ".inst 0xa1402aa0  // ld1h { z0.h, z8.h }, pn10.b/Z, [x21]\n"
    "trn1 z29.h, z18.h, z16.h\n"
    "ld1sb { z16.s }, p0/Z, [x19]\n"
    "addvl x20, SP, #18\n"
    "add z16.h, p0/M, z16.h, z25.h\n"
    ".inst 0xc1687768  // sdot za.s[x11, 0], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc1607769  // sdot za.s[x11, 1], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402a80  // ld1h { z0.h, z8.h }, pn10.b/Z, [x20]\n"
    "addvl x19, SP, #24\n"
    "trn1 z30.h, z17.h, z16.h\n"
    ".inst 0xa0412aa4  // ld1h { z4.h-z5.h }, pn10.b/Z, [x21, #0x2, MUL VL]\n"
    ".inst 0xc168776a  // sdot za.s[x11, 2], { z27.h-z28.h }, z8.h\n"
    "add x15, x15, %x[ld_in_col]\n"
    ".inst 0xc160776b  // sdot za.s[x11, 3], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402a60  // ld1h { z0.h, z8.h }, pn10.b/Z, [x19]\n"
    ".inst 0xc1657788  // sdot za.s[x11, 0], { z28.h-z29.h }, z5.h\n"
    ".inst 0xa0422aaa  // ld1h { z10.h-z11.h }, pn10.b/Z, [x21, #0x4, MUL VL]\n"
    ".inst 0xc1647789  // sdot za.s[x11, 1], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412a84  // ld1h { z4.h-z5.h }, pn10.b/Z, [x20, #0x2, MUL VL]\n"
    ".inst 0xc168776c  // sdot za.s[x11, 4], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc160776d  // sdot za.s[x11, 5], { z27.h-z28.h }, z0.h\n"
    ".inst 0xc165778a  // sdot za.s[x11, 2], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778b  // sdot za.s[x11, 3], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412a64  // ld1h { z4.h-z5.h }, pn10.b/Z, [x19, #0x2, MUL VL]\n"
    ".inst 0xc16b77a8  // sdot za.s[x11, 0], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77a9  // sdot za.s[x11, 1], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422a8a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x20, #0x4, MUL VL]\n"
    ".inst 0xc165778c  // sdot za.s[x11, 4], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778d  // sdot za.s[x11, 5], { z28.h-z29.h }, z4.h\n"
    ".inst 0xc16b77aa  // sdot za.s[x11, 2], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ab  // sdot za.s[x11, 3], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422a6a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x19, #0x4, MUL VL]\n"
    ".inst 0xc16b77ac  // sdot za.s[x11, 4], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ad  // sdot za.s[x11, 5], { z29.h-z30.h }, z10.h\n"
    "18:"  // Padded: 1 priming loads
    "mov x12, #0x0\n"
    ".inst 0x25304500  // psel p0.s, p1.s/Z, p8.s[w12]\n"
    "ld1sb { z19.s }, p0/Z, [x15]\n"
    "add z19.h, p0/M, z19.h, z25.h\n"
    "add x19, x15, %x[ld_in_row]\n"
    ".inst 0x25704500  // psel p0.s, p1.s/Z, p8.s[w12, #1]\n"
    "ld1sb { z18.s }, p0/Z, [x19]\n"
    "add z18.h, p0/M, z18.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25b04500  // psel p0.s, p1.s/Z, p8.s[w12, #2]\n"
    "ld1sb { z17.s }, p0/Z, [x19]\n"
    "add z17.h, p0/M, z17.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25f04500  // psel p0.s, p1.s/Z, p8.s[w12, #3]\n"
    "ld1sb { z16.s }, p0/Z, [x19]\n"
    "add z16.h, p0/M, z16.h, z25.h\n"
    "mov x12, #0x4\n"
    "add x19, x19, %x[ld_in_row]\n"
    "trn1 z27.h, z19.h, z18.h\n"
    "trn1 z28.h, z17.h, z16.h\n"
    ".inst 0x25304500  // psel p0.s, p1.s/Z, p8.s[w12]\n"
    "ld1sb { z18.s }, p0/Z, [x19]\n"
    "add z18.h, p0/M, z18.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25704500  // psel p0.s, p1.s/Z, p8.s[w12, #1]\n"
    "ld1sb { z16.s }, p0/Z, [x19]\n"
    "add z16.h, p0/M, z16.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25b04500  // psel p0.s, p1.s/Z, p8.s[w12, #2]\n"
    "ld1sb { z17.s }, p0/Z, [x19]\n"
    "addvl x22, SP, #6\n"
    "add z17.h, p0/M, z17.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25f04500  // psel p0.s, p1.s/Z, p8.s[w12, #3]\n"
    ".inst 0xa1402ac0  // ld1h { z0.h, z8.h }, pn10.b/Z, [x22]\n"
    "trn1 z29.h, z18.h, z16.h\n"
    "ld1sb { z16.s }, p0/Z, [x19]\n"
    "addvl x21, SP, #12\n"
    "add z16.h, p0/M, z16.h, z25.h\n"
    ".inst 0xc1687768  // sdot za.s[x11, 0], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc1607769  // sdot za.s[x11, 1], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402aa0  // ld1h { z0.h, z8.h }, pn10.b/Z, [x21]\n"
    "addvl x20, SP, #18\n"
    "trn1 z30.h, z17.h, z16.h\n"
    ".inst 0xa0412ac4  // ld1h { z4.h-z5.h }, pn10.b/Z, [x22, #0x2, MUL VL]\n"
    ".inst 0xc168776a  // sdot za.s[x11, 2], { z27.h-z28.h }, z8.h\n"
    "addvl x19, SP, #24\n"
    "add x15, x15, %x[ld_in_col]\n"
    ".inst 0xc160776b  // sdot za.s[x11, 3], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402a80  // ld1h { z0.h, z8.h }, pn10.b/Z, [x20]\n"
    ".inst 0xc1657788  // sdot za.s[x11, 0], { z28.h-z29.h }, z5.h\n"
    ".inst 0xa0422aca  // ld1h { z10.h-z11.h }, pn10.b/Z, [x22, #0x4, MUL VL]\n"
    ".inst 0xc1647789  // sdot za.s[x11, 1], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412aa4  // ld1h { z4.h-z5.h }, pn10.b/Z, [x21, #0x2, MUL VL]\n"
    ".inst 0xc168776c  // sdot za.s[x11, 4], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc160776d  // sdot za.s[x11, 5], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402a60  // ld1h { z0.h, z8.h }, pn10.b/Z, [x19]\n"
    ".inst 0xc165778a  // sdot za.s[x11, 2], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778b  // sdot za.s[x11, 3], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412a84  // ld1h { z4.h-z5.h }, pn10.b/Z, [x20, #0x2, MUL VL]\n"
    ".inst 0xc16b77a8  // sdot za.s[x11, 0], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77a9  // sdot za.s[x11, 1], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422aaa  // ld1h { z10.h-z11.h }, pn10.b/Z, [x21, #0x4, MUL VL]\n"
    ".inst 0xc168776e  // sdot za.s[x11, 6], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc160776f  // sdot za.s[x11, 7], { z27.h-z28.h }, z0.h\n"
    ".inst 0xc165778c  // sdot za.s[x11, 4], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778d  // sdot za.s[x11, 5], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412a64  // ld1h { z4.h-z5.h }, pn10.b/Z, [x19, #0x2, MUL VL]\n"
    ".inst 0xc16b77aa  // sdot za.s[x11, 2], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ab  // sdot za.s[x11, 3], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422a8a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x20, #0x4, MUL VL]\n"
    ".inst 0xc165778e  // sdot za.s[x11, 6], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778f  // sdot za.s[x11, 7], { z28.h-z29.h }, z4.h\n"
    ".inst 0xc16b77ac  // sdot za.s[x11, 4], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ad  // sdot za.s[x11, 5], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422a6a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x19, #0x4, MUL VL]\n"
    ".inst 0xc16b77ae  // sdot za.s[x11, 6], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77af  // sdot za.s[x11, 7], { z29.h-z30.h }, z10.h\n"
    "19:"  // Padded: 0 priming loads
    ".inst 0xa1402be0  // ld1h { z0.h, z8.h }, pn10.b/Z, [SP]\n"
    ".inst 0xa0412be4  // ld1h { z4.h-z5.h }, pn10.b/Z, [SP, #0x2, MUL VL]\n"
    ".inst 0xa0422bea  // ld1h { z10.h-z11.h }, pn10.b/Z, [SP, #0x4, MUL VL]\n"
    "cbz x16, 22f\n"
    "mov x12, #0x0\n"
    ".inst 0x25304500  // psel p0.s, p1.s/Z, p8.s[w12]\n"
    "ld1sb { z19.s }, p0/Z, [x15]\n"
    "add z19.h, p0/M, z19.h, z25.h\n"
    "add x19, x15, %x[ld_in_row]\n"
    ".inst 0x25704500  // psel p0.s, p1.s/Z, p8.s[w12, #1]\n"
    "ld1sb { z18.s }, p0/Z, [x19]\n"
    "add z18.h, p0/M, z18.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25b04500  // psel p0.s, p1.s/Z, p8.s[w12, #2]\n"
    "ld1sb { z17.s }, p0/Z, [x19]\n"
    "add z17.h, p0/M, z17.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25f04500  // psel p0.s, p1.s/Z, p8.s[w12, #3]\n"
    "ld1sb { z16.s }, p0/Z, [x19]\n"
    "add z16.h, p0/M, z16.h, z25.h\n"
    "mov x12, #0x4\n"
    "add x19, x19, %x[ld_in_row]\n"
    "trn1 z27.h, z19.h, z18.h\n"
    "trn1 z28.h, z17.h, z16.h\n"
    ".inst 0x25304500  // psel p0.s, p1.s/Z, p8.s[w12]\n"
    "ld1sb { z19.s }, p0/Z, [x19]\n"
    "add z19.h, p0/M, z19.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25704500  // psel p0.s, p1.s/Z, p8.s[w12, #1]\n"
    "ld1sb { z18.s }, p0/Z, [x19]\n"
    "add z18.h, p0/M, z18.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25b04500  // psel p0.s, p1.s/Z, p8.s[w12, #2]\n"
    "ld1sb { z17.s }, p0/Z, [x19]\n"
    "add z17.h, p0/M, z17.h, z25.h\n"
    "add x19, x19, %x[ld_in_row]\n"
    ".inst 0x25f04500  // psel p0.s, p1.s/Z, p8.s[w12, #3]\n"
    "ld1sb { z16.s }, p0/Z, [x19]\n"
    "add z16.h, p0/M, z16.h, z25.h\n"
    "sub x16, x16, #0x1\n"
    "sub x14, x14, #0x1\n"
    "cmp x16, x14\n"
    "trn1 z29.h, z19.h, z18.h\n"
    "trn1 z30.h, z17.h, z16.h\n"
    "csel x24, x16, x14, LT\n"
    "add x15, x15, %x[ld_in_col]\n"
    "sub x14, x14, x24\n"
    "cbz x24, 21f\n"
    "20:"  // Padded: Main loop
    "mov x12, #0x0\n"
    ".inst 0x25304500  // psel p0.s, p1.s/Z, p8.s[w12]\n"
    "ld1sb { z23.s }, p0/Z, [x15]\n"
    "add z23.h, p0/M, z23.h, z25.h\n"
    "add x23, x15, %x[ld_in_row]\n"
    ".inst 0x25704500  // psel p0.s, p1.s/Z, p8.s[w12, #1]\n"
    "ld1sb { z22.s }, p0/Z, [x23]\n"
    ".inst 0xc1687768  // sdot za.s[x11, 0], { z27.h-z28.h }, z8.h\n"
    "addvl x22, SP, #6\n"
    ".inst 0xc1607769  // sdot za.s[x11, 1], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402ac0  // ld1h { z0.h, z8.h }, pn10.b/Z, [x22]\n"
    "addvl x21, SP, #12\n"
    "add z22.h, p0/M, z22.h, z25.h\n"
    "add x23, x23, %x[ld_in_row]\n"
    ".inst 0x25b04500  // psel p0.s, p1.s/Z, p8.s[w12, #2]\n"
    ".inst 0xc168776a  // sdot za.s[x11, 2], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc160776b  // sdot za.s[x11, 3], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402aa0  // ld1h { z0.h, z8.h }, pn10.b/Z, [x21]\n"
    "addvl x20, SP, #18\n"
    "addvl x19, SP, #24\n"
    "ld1sb { z21.s }, p0/Z, [x23]\n"
    ".inst 0xc1657788  // sdot za.s[x11, 0], { z28.h-z29.h }, z5.h\n"
    "add z21.h, p0/M, z21.h, z25.h\n"
    ".inst 0x25f04500  // psel p0.s, p1.s/Z, p8.s[w12, #3]\n"
    ".inst 0xc1647789  // sdot za.s[x11, 1], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412ac4  // ld1h { z4.h-z5.h }, pn10.b/Z, [x22, #0x2, MUL VL]\n"
    "mov x12, #0x4\n"
    "add x23, x23, %x[ld_in_row]\n"
    ".inst 0xc168776c  // sdot za.s[x11, 4], { z27.h-z28.h }, z8.h\n"
    "ld1sb { z20.s }, p0/Z, [x23]\n"
    "add z20.h, p0/M, z20.h, z25.h\n"
    "add x23, x23, %x[ld_in_row]\n"
    ".inst 0xc160776d  // sdot za.s[x11, 5], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402a80  // ld1h { z0.h, z8.h }, pn10.b/Z, [x20]\n"
    ".inst 0x25304500  // psel p0.s, p1.s/Z, p8.s[w12]\n"
    "subs x24, x24, #0x1\n"
    ".inst 0xc165778a  // sdot za.s[x11, 2], { z28.h-z29.h }, z5.h\n"
    "ld1sb { z19.s }, p0/Z, [x23]\n"
    "add z19.h, p0/M, z19.h, z25.h\n"
    "add x23, x23, %x[ld_in_row]\n"
    ".inst 0xc164778b  // sdot za.s[x11, 3], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412aa4  // ld1h { z4.h-z5.h }, pn10.b/Z, [x21, #0x2, MUL VL]\n"
    ".inst 0x25704500  // psel p0.s, p1.s/Z, p8.s[w12, #1]\n"
    "add x15, x15, %x[ld_in_col]\n"
    ".inst 0xc16b77a8  // sdot za.s[x11, 0], { z29.h-z30.h }, z11.h\n"
    "ld1sb { z18.s }, p0/Z, [x23]\n"
    "add z18.h, p0/M, z18.h, z25.h\n"
    "add x23, x23, %x[ld_in_row]\n"
    ".inst 0xc16a77a9  // sdot za.s[x11, 1], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422aca  // ld1h { z10.h-z11.h }, pn10.b/Z, [x22, #0x4, MUL VL]\n"
    ".inst 0x25b04500  // psel p0.s, p1.s/Z, p8.s[w12, #2]\n"
    ".inst 0xc168776e  // sdot za.s[x11, 6], { z27.h-z28.h }, z8.h\n"
    "ld1sb { z17.s }, p0/Z, [x23]\n"
    "add z17.h, p0/M, z17.h, z25.h\n"
    "add x23, x23, %x[ld_in_row]\n"
    ".inst 0xc160776f  // sdot za.s[x11, 7], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402a60  // ld1h { z0.h, z8.h }, pn10.b/Z, [x19]\n"
    ".inst 0x25f04500  // psel p0.s, p1.s/Z, p8.s[w12, #3]\n"
    ".inst 0xc165778c  // sdot za.s[x11, 4], { z28.h-z29.h }, z5.h\n"
    "ld1sb { z16.s }, p0/Z, [x23]\n"
    "add z16.h, p0/M, z16.h, z25.h\n"
    ".inst 0xc164778d  // sdot za.s[x11, 5], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412a84  // ld1h { z4.h-z5.h }, pn10.b/Z, [x20, #0x2, MUL VL]\n"
    ".inst 0xc16b77aa  // sdot za.s[x11, 2], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ab  // sdot za.s[x11, 3], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422aaa  // ld1h { z10.h-z11.h }, pn10.b/Z, [x21, #0x4, MUL VL]\n"
    ".inst 0xc165778e  // sdot za.s[x11, 6], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778f  // sdot za.s[x11, 7], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412a64  // ld1h { z4.h-z5.h }, pn10.b/Z, [x19, #0x2, MUL VL]\n"
    ".inst 0xc16b77ac  // sdot za.s[x11, 4], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ad  // sdot za.s[x11, 5], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422a8a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x20, #0x4, MUL VL]\n"
    ".inst 0xc16b77ae  // sdot za.s[x11, 6], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77af  // sdot za.s[x11, 7], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422a6a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x19, #0x4, MUL VL]\n"
    ".inst 0xc1681768  // sdot za.s[x8, 0], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc1601769  // sdot za.s[x8, 1], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402be0  // ld1h { z0.h, z8.h }, pn10.b/Z, [SP]\n"
    "trn1 z27.h, z23.h, z22.h\n"
    ".inst 0xc1651788  // sdot za.s[x8, 0], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc1641789  // sdot za.s[x8, 1], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412be4  // ld1h { z4.h-z5.h }, pn10.b/Z, [SP, #0x2, MUL VL]\n"
    "trn1 z28.h, z21.h, z20.h\n"
    ".inst 0xc16b17a8  // sdot za.s[x8, 0], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a17a9  // sdot za.s[x8, 1], { z29.h-z30.h }, z10.h\n"
    "add x8, x8, #0x2\n"
    ".inst 0xa0422bea  // ld1h { z10.h-z11.h }, pn10.b/Z, [SP, #0x4, MUL VL]\n"
    "trn1 z29.h, z19.h, z18.h\n"
    ".inst 0xc006680c  // mova { z12.d-z13.d }, za.d[x11, #0]\n"
    "trn1 z30.h, z17.h, z16.h\n"
    ".inst 0xc006682e  // mova { z14.d-z15.d }, za.d[x11, #1]\n"
    ".inst 0xc1a3ac0c  // sqdmulh { z12.s-z15.s }, { z12.s-z15.s }, z3.s\n"
    "add x11, x11, #0x2\n"
    ".inst 0xc1a1aa2c  // srshl { z12.s-z15.s }, { z12.s-z15.s }, z1.s\n"
    ".inst 0xc00408c0  // mova za.d[x8, #0], { z6.d-z7.d }\n"
    ".inst 0xc1a9ab0c  // add { z12.s-z15.s }, { z12.s-z15.s }, z9.s\n"
    ".inst 0xc00408c1  // mova za.d[x8, #1], { z6.d-z7.d }\n"
    ".inst 0xc1bfcf0c  // sclamp { z12.s-z15.s }, z24.s, z31.s\n"
    "st1b { z12.s }, p1, [x13]\n"
    "add x13, x13, x10\n"
    "st1b { z14.s }, p1, [x4]\n"
    "add x4, x4, x9\n"
    "st1b { z13.s }, p1, [x28]\n"
    "add x28, x28, x26\n"
    "st1b { z15.s }, p1, [x27]\n"
    "add x27, x27, x25\n"
    "bgt 20b\n"
    "21:"  // Main loop tail
    "addvl x22, SP, #6\n"
    ".inst 0xc1687768  // sdot za.s[x11, 0], { z27.h-z28.h }, z8.h\n"
    "addvl x21, SP, #12\n"
    ".inst 0xc1607769  // sdot za.s[x11, 1], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402ac0  // ld1h { z0.h, z8.h }, pn10.b/Z, [x22]\n"
    "addvl x20, SP, #18\n"
    "addvl x19, SP, #24\n"
    ".inst 0xc168776a  // sdot za.s[x11, 2], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc160776b  // sdot za.s[x11, 3], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402aa0  // ld1h { z0.h, z8.h }, pn10.b/Z, [x21]\n"
    ".inst 0xc1657788  // sdot za.s[x11, 0], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc1647789  // sdot za.s[x11, 1], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412ac4  // ld1h { z4.h-z5.h }, pn10.b/Z, [x22, #0x2, MUL VL]\n"
    ".inst 0xc168776c  // sdot za.s[x11, 4], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc160776d  // sdot za.s[x11, 5], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402a80  // ld1h { z0.h, z8.h }, pn10.b/Z, [x20]\n"
    ".inst 0xc165778a  // sdot za.s[x11, 2], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778b  // sdot za.s[x11, 3], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412aa4  // ld1h { z4.h-z5.h }, pn10.b/Z, [x21, #0x2, MUL VL]\n"
    ".inst 0xc16b77a8  // sdot za.s[x11, 0], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77a9  // sdot za.s[x11, 1], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422aca  // ld1h { z10.h-z11.h }, pn10.b/Z, [x22, #0x4, MUL VL]\n"
    ".inst 0xc168776e  // sdot za.s[x11, 6], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc160776f  // sdot za.s[x11, 7], { z27.h-z28.h }, z0.h\n"
    ".inst 0xa1402a60  // ld1h { z0.h, z8.h }, pn10.b/Z, [x19]\n"
    ".inst 0xc165778c  // sdot za.s[x11, 4], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778d  // sdot za.s[x11, 5], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412a84  // ld1h { z4.h-z5.h }, pn10.b/Z, [x20, #0x2, MUL VL]\n"
    ".inst 0xc16b77aa  // sdot za.s[x11, 2], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ab  // sdot za.s[x11, 3], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422aaa  // ld1h { z10.h-z11.h }, pn10.b/Z, [x21, #0x4, MUL VL]\n"
    ".inst 0xc165778e  // sdot za.s[x11, 6], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc164778f  // sdot za.s[x11, 7], { z28.h-z29.h }, z4.h\n"
    ".inst 0xa0412a64  // ld1h { z4.h-z5.h }, pn10.b/Z, [x19, #0x2, MUL VL]\n"
    ".inst 0xc16b77ac  // sdot za.s[x11, 4], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77ad  // sdot za.s[x11, 5], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422a8a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x20, #0x4, MUL VL]\n"
    ".inst 0xc16b77ae  // sdot za.s[x11, 6], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a77af  // sdot za.s[x11, 7], { z29.h-z30.h }, z10.h\n"
    ".inst 0xa0422a6a  // ld1h { z10.h-z11.h }, pn10.b/Z, [x19, #0x4, MUL VL]\n"
    ".inst 0xc1681768  // sdot za.s[x8, 0], { z27.h-z28.h }, z8.h\n"
    ".inst 0xc1601769  // sdot za.s[x8, 1], { z27.h-z28.h }, z0.h\n"
    ".inst 0xc1651788  // sdot za.s[x8, 0], { z28.h-z29.h }, z5.h\n"
    ".inst 0xc1641789  // sdot za.s[x8, 1], { z28.h-z29.h }, z4.h\n"
    ".inst 0xc16b17a8  // sdot za.s[x8, 0], { z29.h-z30.h }, z11.h\n"
    ".inst 0xc16a17a9  // sdot za.s[x8, 1], { z29.h-z30.h }, z10.h\n"
    "add x8, x8, #0x2\n"
    ".inst 0xc006680c  // mova { z12.d-z13.d }, za.d[x11, #0]\n"
    ".inst 0xc006682e  // mova { z14.d-z15.d }, za.d[x11, #1]\n"
    ".inst 0xc1a3ac0c  // sqdmulh { z12.s-z15.s }, { z12.s-z15.s }, z3.s\n"
    "add x11, x11, #0x2\n"
    ".inst 0xc1a1aa2c  // srshl { z12.s-z15.s }, { z12.s-z15.s }, z1.s\n"
    ".inst 0xc00408c0  // mova za.d[x8, #0], { z6.d-z7.d }\n"
    ".inst 0xc1a9ab0c  // add { z12.s-z15.s }, { z12.s-z15.s }, z9.s\n"
    ".inst 0xc00408c1  // mova za.d[x8, #1], { z6.d-z7.d }\n"
    ".inst 0xc1bfcf0c  // sclamp { z12.s-z15.s }, z24.s, z31.s\n"
    "st1b { z12.s }, p1, [x13]\n"
    "add x13, x13, x10\n"
    "st1b { z14.s }, p1, [x4]\n"
    "add x4, x4, x9\n"
    "st1b { z13.s }, p1, [x28]\n"
    "add x28, x28, x26\n"
    "st1b { z15.s }, p1, [x27]\n"
    "add x27, x27, x25\n"
    "22:"  // Main loop skip tail
    "cbz x14, 24f\n"
    "23:"  // Right padding loop
    ".inst 0xc006680c  // mova { z12.d-z13.d }, za.d[x11, #0]\n"
    "add x8, x8, #0x2\n"
    "subs x14, x14, #0x1\n"
    ".inst 0xc006682e  // mova { z14.d-z15.d }, za.d[x11, #1]\n"
    ".inst 0xc1a3ac0c  // sqdmulh { z12.s-z15.s }, { z12.s-z15.s }, z3.s\n"
    "add x11, x11, #0x2\n"
    ".inst 0xc1a1aa2c  // srshl { z12.s-z15.s }, { z12.s-z15.s }, z1.s\n"
    ".inst 0xc00408c0  // mova za.d[x8, #0], { z6.d-z7.d }\n"
    ".inst 0xc1a9ab0c  // add { z12.s-z15.s }, { z12.s-z15.s }, z9.s\n"
    ".inst 0xc00408c1  // mova za.d[x8, #1], { z6.d-z7.d }\n"
    ".inst 0xc1bfcf0c  // sclamp { z12.s-z15.s }, z24.s, z31.s\n"
    "st1b { z12.s }, p1, [x13]\n"
    "add x13, x13, x10\n"
    "st1b { z14.s }, p1, [x4]\n"
    "add x4, x4, x9\n"
    "st1b { z13.s }, p1, [x28]\n"
    "add x28, x28, x26\n"
    "st1b { z15.s }, p1, [x27]\n"
    "add x27, x27, x25\n"
    "bgt 23b\n"
    "24:"  // End
    "ldr x23, [%x[args], %[offsetof_Args_weights]]\n"
    "incw x23, ALL, MUL #16\n"
    "incw x23, ALL, MUL #9\n"
    "str x23, [%x[args], %[offsetof_Args_weights]]\n"
    "ldr x19, [%x[args], %[offsetof_Args_ld_in_vl]]\n"
    "incw x17\n"
    "whilelt p1.s, x17, x7\n"
    "ldr x15, [%x[args], %[offsetof_Args_inptr]]\n"
    "add x15, x15, x19\n"
    "str x15, [%x[args], %[offsetof_Args_inptr]]\n"
    "ldr x24, [%x[args], %[offsetof_Args_outptrs]]\n"
    "ldr x23, [%x[args], %[offsetof_Args_ld_out_vls]]\n"
    "ldp x22, x21, [x24, #0x0]\n"
    "ldp x20, x19, [x23, #0x0]\n"
    "add x22, x22, x20\n"
    "add x21, x21, x19\n"
    "stp x22, x21, [x24, #0x0]\n"
    "ldp x22, x21, [x24, #0x10]\n"
    "ldp x20, x19, [x23, #0x10]\n"
    "add x22, x22, x20\n"
    "add x21, x21, x19\n"
    "stp x22, x21, [x24, #0x10]\n"
    "b.any 1b\n"
    "addvl SP, SP, #30\n"
    ".inst 0xd503467f  // SMSTOP\n"
    :
    : [args] "r" (&args), [ld_in_col] "r" (ld_in_col), [ld_in_row] "r" (ld_in_row), [offsetof_Args_current_channel] "I" (offsetof(Args, current_channel)), [offsetof_Args_inptr] "I" (offsetof(Args, inptr)), [offsetof_Args_input_cols] "I" (offsetof(Args, input_cols)), [offsetof_Args_ld_in_vl] "I" (offsetof(Args, ld_in_vl)), [offsetof_Args_ld_out_cols] "I" (offsetof(Args, ld_out_cols)), [offsetof_Args_ld_out_vls] "I" (offsetof(Args, ld_out_vls)), [offsetof_Args_n_channels] "I" (offsetof(Args, n_channels)), [offsetof_Args_outptrs] "I" (offsetof(Args, outptrs)), [offsetof_Args_output_cols] "I" (offsetof(Args, output_cols)), [offsetof_Args_pad_bottom] "I" (offsetof(Args, pad_bottom)), [offsetof_Args_pad_left] "I" (offsetof(Args, pad_left)), [offsetof_Args_pad_top] "I" (offsetof(Args, pad_top)), [offsetof_Args_weights] "I" (offsetof(Args, weights)), [offsetof_Requantize32_a_offset] "I" (offsetof(arm_gemm::Requantize32, a_offset)), [offsetof_Requantize32_b_offset] "I" (offsetof(arm_gemm::Requantize32, b_offset)), [offsetof_Requantize32_bias] "I" (offsetof(arm_gemm::Requantize32, bias)), [offsetof_Requantize32_c_offset] "I" (offsetof(arm_gemm::Requantize32, c_offset)), [offsetof_Requantize32_maxval] "I" (offsetof(arm_gemm::Requantize32, maxval)), [offsetof_Requantize32_minval] "I" (offsetof(arm_gemm::Requantize32, minval)), [offsetof_Requantize32_per_channel_muls] "I" (offsetof(arm_gemm::Requantize32, per_channel_muls)), [offsetof_Requantize32_per_channel_right_shifts] "I" (offsetof(arm_gemm::Requantize32, per_channel_right_shifts)), [offsetof_Requantize32_per_layer_mul] "I" (offsetof(arm_gemm::Requantize32, per_layer_mul)), [offsetof_Requantize32_per_layer_right_shift] "I" (offsetof(arm_gemm::Requantize32, per_layer_right_shift)), [qp] "r" (&qp)
    : "cc", "memory", "p0", "p1", "p2", "p3", "p4", "p5", "p6", "p7", "p8", "p9", "p10", "p11", "p12", "p13", "p14", "p15", "x4", "x5", "x6", "x7", "x8", "x9", "x10", "x11", "x12", "x13", "x14", "x15", "x16", "x17", "x19", "x20", "x21", "x22", "x23", "x24", "x25", "x26", "x27", "x28", "z0", "z1", "z2", "z3", "z4", "z5", "z6", "z7", "z8", "z9", "z10", "z11", "z12", "z13", "z14", "z15", "z16", "z17", "z18", "z19", "z20", "z21", "z22", "z23", "z24", "z25", "z26", "z27", "z28", "z29", "z30", "z31"
  );
}

}  // namespace depthwise
}  // namespace arm_conv

#endif  // defined(ARM_COMPUTE_ENABLE_SME2)
