/*
 * Copyright (c) 2021-2022 Arm Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include "arm_gemm.hpp"

#include <cstddef>
#include <cstdint>

#if defined(__aarch64__) && defined(ARM_COMPUTE_ENABLE_SVE)

namespace arm_conv {
namespace depthwise {

void sve_s8q_nhwc_5x5_s1_output2x2_mla_depthfirst_impl(
  const unsigned int n_channels,
  const int8_t *const *const inptrs,
  const int8_t *const weights,
  const int32_t *const bias,
  const arm_gemm::Requantize32 &qp,
  const int32_t *const requant_muls,
  const int32_t *const requant_shifts,
  int8_t *const *const outptrs
)
{
  struct Params
  {
    long unsigned int n_channels;
    const int8_t *weights;
    const int32_t *bias;
    const arm_gemm::Requantize32 *requant;
    const int32_t *const requant_muls;
    const int32_t *const requant_shifts;
    int8_t *const *const outptrs;
    const int8_t *inptrs[36];

    Params(
      long unsigned int n_channels,
      const int8_t *const *inptrs_raw,
      const int8_t *const weights,
      const int32_t *const bias,
      const arm_gemm::Requantize32 &qp,
      const int32_t *const requant_muls,
      const int32_t *const requant_shifts,
      int8_t *const *outptrs
    ) : n_channels(n_channels), weights(weights), bias(bias),
        requant(&qp), requant_muls(requant_muls),
        requant_shifts(requant_shifts), outptrs(outptrs)
    {
      inptrs[0] = inptrs_raw[0];
      inptrs[1] = inptrs_raw[1];
      inptrs[2] = inptrs_raw[6];
      inptrs[3] = inptrs_raw[7];
      inptrs[4] = inptrs_raw[2];
      inptrs[5] = inptrs_raw[8];
      inptrs[6] = inptrs_raw[3];
      inptrs[7] = inptrs_raw[4];
      inptrs[8] = inptrs_raw[11];
      inptrs[9] = inptrs_raw[12];
      inptrs[10] = inptrs_raw[9];
      inptrs[11] = inptrs_raw[10];
      inptrs[12] = inptrs_raw[5];
      inptrs[13] = inptrs_raw[13];
      inptrs[14] = inptrs_raw[14];
      inptrs[15] = inptrs_raw[15];
      inptrs[16] = inptrs_raw[16];
      inptrs[17] = inptrs_raw[17];
      inptrs[18] = inptrs_raw[18];
      inptrs[19] = inptrs_raw[19];
      inptrs[20] = inptrs_raw[20];
      inptrs[21] = inptrs_raw[21];
      inptrs[22] = inptrs_raw[22];
      inptrs[23] = inptrs_raw[23];
      inptrs[24] = inptrs_raw[24];
      inptrs[25] = inptrs_raw[25];
      inptrs[26] = inptrs_raw[26];
      inptrs[27] = inptrs_raw[27];
      inptrs[28] = inptrs_raw[28];
      inptrs[29] = inptrs_raw[29];
      inptrs[30] = inptrs_raw[30];
      inptrs[31] = inptrs_raw[31];
      inptrs[32] = inptrs_raw[32];
      inptrs[33] = inptrs_raw[33];
      inptrs[34] = inptrs_raw[34];
      inptrs[35] = inptrs_raw[35];

    }
  };

  const Params params(n_channels, inptrs, weights, bias, qp,
                      requant_muls, requant_shifts, outptrs);

  __asm__ __volatile__(
    "ldr x0, [%x[params], %[offsetof_Params_n_channels]]\n"
    "ptrue p4.b\n"
    "ldr x1, [%x[params], %[offsetof_Params_weights]]\n"
    "mov x2, #0x0\n"
    "ldr x22, [%x[params], %[offsetof_Params_requant]]\n"
    "mov x3, #0x0\n"
    "ldr x4, [%x[params], %[offsetof_Params_requant_muls]]\n"
    "add x5, %x[params], %[offsetof_Params_inptrs]\n"
    "ldr x6, [%x[params], %[offsetof_Params_requant_shifts]]\n"
    "add x19, x22, %[offsetof_Requantize32_a_offset]\n"
    "ldr x21, [%x[params], %[offsetof_Params_outptrs]]\n"
    "add x20, x22, %[offsetof_Requantize32_b_offset]\n"
    "ld1rb { z17.b }, p4/Z, [x19]\n"
    "add x19, x22, %[offsetof_Requantize32_c_offset]\n"
    "ld1rb { z13.b }, p4/Z, [x20]\n"
    "add x20, x22, %[offsetof_Requantize32_minval]\n"
    "ld1rw { z14.s }, p4/Z, [x19]\n"
    "add x19, x22, %[offsetof_Requantize32_maxval]\n"
    "ld1rw { z5.s }, p4/Z, [x20]\n"
    "whilelt p3.h, x2, x0\n"
    "ld1rw { z15.s }, p4/Z, [x19]\n"
    "whilelt p2.s, x2, x0\n"
    "ldp x7, x8, [x21, #0x0]\n"
    "mov x19, x2\n"
    "incw x19\n"
    "ldp x17, x16, [x21, #0x10]\n"
    "whilelt p1.s, x19, x0\n"
    "ldr x19, [%x[params], %[offsetof_Params_bias]]\n"
    "ld1w { z19.s }, p2/Z, [x19]\n"
    "ld1w { z6.s }, p1/Z, [x19, #1, MUL VL]\n"
    "uzp1 z11.s, z19.s, z6.s\n"
    "addvl x19, x19, #2\n"
    "str x19, [%x[params], %[offsetof_Params_bias]]\n"
    "uzp2 z16.s, z19.s, z6.s\n"
    "mov z19.d, z11.d\n"
    "ld1sb { z0.h }, p4/Z, [x1]\n"
    ".inst 0x454d1000  // ssublb z0.h, z0.b, z13.b\n"
    "mov z9.d, z16.d\n"
    "ld1sb { z1.h }, p4/Z, [x1, #1, MUL VL]\n"
    "mov z7.d, z11.d\n"
    "ld1sb { z2.h }, p4/Z, [x1, #2, MUL VL]\n"
    ".inst 0x454d1021  // ssublb z1.h, z1.b, z13.b\n"
    "mov z6.d, z16.d\n"
    "ld1sb { z3.h }, p4/Z, [x1, #3, MUL VL]\n"
    "mov z12.d, z11.d\n"
    "ld1sb { z4.h }, p4/Z, [x1, #4, MUL VL]\n"
    ".inst 0x454d1042  // ssublb z2.h, z2.b, z13.b\n"
    "mov z8.d, z16.d\n"
    "ldp x28, x27, [x5, #0x0]\n"
    ".inst 0x454d1063  // ssublb z3.h, z3.b, z13.b\n"
    "ldp x26, x25, [x5, #0x10]\n"
    ".inst 0x454d1084  // ssublb z4.h, z4.b, z13.b\n"
    "ldp x24, x23, [x5, #0x20]\n"
    "ldp x22, x21, [x5, #0x30]\n"
    "ldp x20, x19, [x5, #0x40]\n"
    "ld1sb { z31.h }, p3/Z, [x28, x2]\n"
    ".inst 0x455113ff  // ssublb z31.h, z31.b, z17.b\n"
    "ld1sb { z30.h }, p3/Z, [x27, x2]\n"
    "ld1sb { z29.h }, p3/Z, [x26, x2]\n"
    ".inst 0x455113de  // ssublb z30.h, z30.b, z17.b\n"
    "ld1sb { z28.h }, p3/Z, [x25, x2]\n"
    "ld1sb { z27.h }, p3/Z, [x24, x2]\n"
    ".inst 0x455113bd  // ssublb z29.h, z29.b, z17.b\n"
    "ld1sb { z23.h }, p3/Z, [x23, x2]\n"
    ".inst 0x4551139c  // ssublb z28.h, z28.b, z17.b\n"
    "ld1sb { z25.h }, p3/Z, [x22, x2]\n"
    "ld1sb { z24.h }, p3/Z, [x21, x2]\n"
    ".inst 0x4551137b  // ssublb z27.h, z27.b, z17.b\n"
    "ld1sb { z26.h }, p3/Z, [x20, x2]\n"
    ".inst 0x455112f7  // ssublb z23.h, z23.b, z17.b\n"
    "ld1sb { z22.h }, p3/Z, [x19, x2]\n"
    ".inst 0x45511339  // ssublb z25.h, z25.b, z17.b\n"
    ".inst 0x45511318  // ssublb z24.h, z24.b, z17.b\n"
    ".inst 0x4551135a  // ssublb z26.h, z26.b, z17.b\n"
    ".inst 0x455112d6  // ssublb z22.h, z22.b, z17.b\n"
    "1:"  // Loop
    ".inst 0x448043eb  // smlalb z11.s, p4/M, z31.h, z0.h\n"
    "ldr x20, [x5, #0x50]\n"
    "whilelt p0.h, x3, x0\n"
    ".inst 0x448047f0  // smlalt z16.s, p4/M, z31.h, z0.h\n"
    "ldr x19, [x5, #0x58]\n"
    ".inst 0x448043d3  // smlalb z19.s, p4/M, z30.h, z0.h\n"
    "ldr x25, [x5, #0x60]\n"
    ".inst 0x448047c9  // smlalt z9.s, p4/M, z30.h, z0.h\n"
    "ld1sb { z31.h }, p3/Z, [x20, x2]\n"
    ".inst 0x455113ff  // ssublb z31.h, z31.b, z17.b\n"
    ".inst 0x448043a7  // smlalb z7.s, p4/M, z29.h, z0.h\n"
    "ldr x24, [x5, #0x68]\n"
    ".inst 0x448047a6  // smlalt z6.s, p4/M, z29.h, z0.h\n"
    "ldr x23, [x5, #0x70]\n"
    ".inst 0x4480438c  // smlalb z12.s, p4/M, z28.h, z0.h\n"
    "ldr x22, [x5, #0x78]\n"
    ".inst 0x44804788  // smlalt z8.s, p4/M, z28.h, z0.h\n"
    "ld1sb { z0.h }, p4/Z, [x1, #5, MUL VL]\n"
    ".inst 0x454d1000  // ssublb z0.h, z0.b, z13.b\n"
    ".inst 0x448143cb  // smlalb z11.s, p4/M, z30.h, z1.h\n"
    "ldr x15, [x5, #0x80]\n"
    ".inst 0x448147d0  // smlalt z16.s, p4/M, z30.h, z1.h\n"
    "ld1sb { z30.h }, p3/Z, [x19, x2]\n"
    ".inst 0x455113de  // ssublb z30.h, z30.b, z17.b\n"
    ".inst 0x44814373  // smlalb z19.s, p4/M, z27.h, z1.h\n"
    "ldr x21, [x5, #0x88]\n"
    ".inst 0x44814769  // smlalt z9.s, p4/M, z27.h, z1.h\n"
    "ldr x20, [x5, #0x90]\n"
    ".inst 0x44814387  // smlalb z7.s, p4/M, z28.h, z1.h\n"
    "ldr x19, [x5, #0x98]\n"
    ".inst 0x44814786  // smlalt z6.s, p4/M, z28.h, z1.h\n"
    "ldr x14, [x5, #0xa0]\n"
    ".inst 0x448142ec  // smlalb z12.s, p4/M, z23.h, z1.h\n"
    "ldr x13, [x5, #0xa8]\n"
    ".inst 0x448146e8  // smlalt z8.s, p4/M, z23.h, z1.h\n"
    "ld1sb { z1.h }, p4/Z, [x1, #6, MUL VL]\n"
    ".inst 0x454d1021  // ssublb z1.h, z1.b, z13.b\n"
    ".inst 0x4482436b  // smlalb z11.s, p4/M, z27.h, z2.h\n"
    "ldr x12, [x5, #0xb0]\n"
    ".inst 0x44824770  // smlalt z16.s, p4/M, z27.h, z2.h\n"
    "ld1sb { z27.h }, p3/Z, [x25, x2]\n"
    ".inst 0x4551137b  // ssublb z27.h, z27.b, z17.b\n"
    ".inst 0x44824333  // smlalb z19.s, p4/M, z25.h, z2.h\n"
    "ldr x11, [x5, #0xb8]\n"
    ".inst 0x44824729  // smlalt z9.s, p4/M, z25.h, z2.h\n"
    "ldr x10, [x5, #0xc0]\n"
    ".inst 0x448242e7  // smlalb z7.s, p4/M, z23.h, z2.h\n"
    "ldr x9, [x5, #0xc8]\n"
    ".inst 0x448246e6  // smlalt z6.s, p4/M, z23.h, z2.h\n"
    "ldr x28, [x5, #0xd0]\n"
    ".inst 0x448243ec  // smlalb z12.s, p4/M, z31.h, z2.h\n"
    "ldr x27, [x5, #0xd8]\n"
    ".inst 0x448247e8  // smlalt z8.s, p4/M, z31.h, z2.h\n"
    "ld1sb { z2.h }, p4/Z, [x1, #7, MUL VL]\n"
    "inch x1, ALL, MUL #8\n"
    ".inst 0x4483432b  // smlalb z11.s, p4/M, z25.h, z3.h\n"
    "ldr x26, [x5, #0xe0]\n"
    ".inst 0x454d1042  // ssublb z2.h, z2.b, z13.b\n"
    ".inst 0x44834730  // smlalt z16.s, p4/M, z25.h, z3.h\n"
    "ld1sb { z25.h }, p3/Z, [x24, x2]\n"
    ".inst 0x44834313  // smlalb z19.s, p4/M, z24.h, z3.h\n"
    "ldr x25, [x5, #0xe8]\n"
    ".inst 0x45511339  // ssublb z25.h, z25.b, z17.b\n"
    ".inst 0x44834709  // smlalt z9.s, p4/M, z24.h, z3.h\n"
    "ld1w { z18.s }, p2/Z, [x4]\n"
    ".inst 0x448343e7  // smlalb z7.s, p4/M, z31.h, z3.h\n"
    "ld1w { z20.s }, p1/Z, [x4, #1, MUL VL]\n"
    "addvl x4, x4, #2\n"
    ".inst 0x448347e6  // smlalt z6.s, p4/M, z31.h, z3.h\n"
    ".inst 0x448343cc  // smlalb z12.s, p4/M, z30.h, z3.h\n"
    ".inst 0x448347c8  // smlalt z8.s, p4/M, z30.h, z3.h\n"
    "ld1sb { z3.h }, p4/Z, [x1]\n"
    ".inst 0x454d1063  // ssublb z3.h, z3.b, z13.b\n"
    "uzp1 z21.s, z18.s, z20.s\n"
    "uzp2 z10.s, z18.s, z20.s\n"
    "ld1w { z18.s }, p2/Z, [x6]\n"
    ".inst 0x4484430b  // smlalb z11.s, p4/M, z24.h, z4.h\n"
    "ld1w { z20.s }, p1/Z, [x6, #1, MUL VL]\n"
    "addvl x6, x6, #2\n"
    ".inst 0x44844710  // smlalt z16.s, p4/M, z24.h, z4.h\n"
    "ld1sb { z24.h }, p3/Z, [x23, x2]\n"
    ".inst 0x45511318  // ssublb z24.h, z24.b, z17.b\n"
    ".inst 0x44844373  // smlalb z19.s, p4/M, z27.h, z4.h\n"
    "ldr x24, [x5, #0xf0]\n"
    ".inst 0x44844769  // smlalt z9.s, p4/M, z27.h, z4.h\n"
    "ld1sb { z27.h }, p3/Z, [x22, x2]\n"
    ".inst 0x4551137b  // ssublb z27.h, z27.b, z17.b\n"
    ".inst 0x448443c7  // smlalb z7.s, p4/M, z30.h, z4.h\n"
    "ldr x23, [x5, #0xf8]\n"
    ".inst 0x448447c6  // smlalt z6.s, p4/M, z30.h, z4.h\n"
    ".inst 0x4484434c  // smlalb z12.s, p4/M, z26.h, z4.h\n"
    ".inst 0x44844748  // smlalt z8.s, p4/M, z26.h, z4.h\n"
    "ld1sb { z4.h }, p4/Z, [x1, #1, MUL VL]\n"
    ".inst 0x454d1084  // ssublb z4.h, z4.b, z13.b\n"
    ".inst 0x448043ab  // smlalb z11.s, p4/M, z29.h, z0.h\n"
    ".inst 0x448047b0  // smlalt z16.s, p4/M, z29.h, z0.h\n"
    "uzp1 z29.s, z18.s, z20.s\n"
    "uzp2 z20.s, z18.s, z20.s\n"
    ".inst 0x44804393  // smlalb z19.s, p4/M, z28.h, z0.h\n"
    ".inst 0x44804789  // smlalt z9.s, p4/M, z28.h, z0.h\n"
    ".inst 0x448042c7  // smlalb z7.s, p4/M, z22.h, z0.h\n"
    ".inst 0x448046c6  // smlalt z6.s, p4/M, z22.h, z0.h\n"
    ".inst 0x4480432c  // smlalb z12.s, p4/M, z25.h, z0.h\n"
    ".inst 0x44804728  // smlalt z8.s, p4/M, z25.h, z0.h\n"
    "ld1sb { z0.h }, p4/Z, [x1, #2, MUL VL]\n"
    ".inst 0x454d1000  // ssublb z0.h, z0.b, z13.b\n"
    ".inst 0x4481438b  // smlalb z11.s, p4/M, z28.h, z1.h\n"
    ".inst 0x44814790  // smlalt z16.s, p4/M, z28.h, z1.h\n"
    "ld1sb { z28.h }, p3/Z, [x21, x2]\n"
    ".inst 0x4551139c  // ssublb z28.h, z28.b, z17.b\n"
    ".inst 0x448142f3  // smlalb z19.s, p4/M, z23.h, z1.h\n"
    "ldr x22, [x5, #0x100]\n"
    ".inst 0x448146e9  // smlalt z9.s, p4/M, z23.h, z1.h\n"
    ".inst 0x44814327  // smlalb z7.s, p4/M, z25.h, z1.h\n"
    ".inst 0x44814726  // smlalt z6.s, p4/M, z25.h, z1.h\n"
    ".inst 0x4481430c  // smlalb z12.s, p4/M, z24.h, z1.h\n"
    ".inst 0x44814708  // smlalt z8.s, p4/M, z24.h, z1.h\n"
    "ld1sb { z1.h }, p4/Z, [x1, #3, MUL VL]\n"
    ".inst 0x454d1021  // ssublb z1.h, z1.b, z13.b\n"
    ".inst 0x448242eb  // smlalb z11.s, p4/M, z23.h, z2.h\n"
    ".inst 0x448246f0  // smlalt z16.s, p4/M, z23.h, z2.h\n"
    "ld1sb { z23.h }, p3/Z, [x15, x2]\n"
    ".inst 0x455112f7  // ssublb z23.h, z23.b, z17.b\n"
    ".inst 0x448243f3  // smlalb z19.s, p4/M, z31.h, z2.h\n"
    "ldr x21, [x5, #0x108]\n"
    ".inst 0x448247e9  // smlalt z9.s, p4/M, z31.h, z2.h\n"
    ".inst 0x44824307  // smlalb z7.s, p4/M, z24.h, z2.h\n"
    ".inst 0x44824706  // smlalt z6.s, p4/M, z24.h, z2.h\n"
    ".inst 0x4482436c  // smlalb z12.s, p4/M, z27.h, z2.h\n"
    ".inst 0x44824768  // smlalt z8.s, p4/M, z27.h, z2.h\n"
    "ld1sb { z2.h }, p4/Z, [x1, #4, MUL VL]\n"
    ".inst 0x454d1042  // ssublb z2.h, z2.b, z13.b\n"
    ".inst 0x448343eb  // smlalb z11.s, p4/M, z31.h, z3.h\n"
    ".inst 0x448347f0  // smlalt z16.s, p4/M, z31.h, z3.h\n"
    "ld1sb { z31.h }, p3/Z, [x20, x2]\n"
    ".inst 0x455113ff  // ssublb z31.h, z31.b, z17.b\n"
    ".inst 0x448343d3  // smlalb z19.s, p4/M, z30.h, z3.h\n"
    "ldr x20, [x5, #0x110]\n"
    ".inst 0x448347c9  // smlalt z9.s, p4/M, z30.h, z3.h\n"
    ".inst 0x44834367  // smlalb z7.s, p4/M, z27.h, z3.h\n"
    ".inst 0x44834766  // smlalt z6.s, p4/M, z27.h, z3.h\n"
    ".inst 0x448342ec  // smlalb z12.s, p4/M, z23.h, z3.h\n"
    ".inst 0x448346e8  // smlalt z8.s, p4/M, z23.h, z3.h\n"
    "ld1sb { z3.h }, p4/Z, [x1, #5, MUL VL]\n"
    ".inst 0x454d1063  // ssublb z3.h, z3.b, z13.b\n"
    ".inst 0x448443cb  // smlalb z11.s, p4/M, z30.h, z4.h\n"
    ".inst 0x448447d0  // smlalt z16.s, p4/M, z30.h, z4.h\n"
    "ld1sb { z30.h }, p3/Z, [x19, x2]\n"
    ".inst 0x455113de  // ssublb z30.h, z30.b, z17.b\n"
    ".inst 0x44844353  // smlalb z19.s, p4/M, z26.h, z4.h\n"
    "ldr x19, [x5, #0x118]\n"
    ".inst 0x44844749  // smlalt z9.s, p4/M, z26.h, z4.h\n"
    "ld1sb { z26.h }, p3/Z, [x14, x2]\n"
    ".inst 0x4551135a  // ssublb z26.h, z26.b, z17.b\n"
    ".inst 0x448442e7  // smlalb z7.s, p4/M, z23.h, z4.h\n"
    ".inst 0x448446e6  // smlalt z6.s, p4/M, z23.h, z4.h\n"
    ".inst 0x4484438c  // smlalb z12.s, p4/M, z28.h, z4.h\n"
    ".inst 0x44844788  // smlalt z8.s, p4/M, z28.h, z4.h\n"
    "ld1sb { z4.h }, p4/Z, [x1, #6, MUL VL]\n"
    ".inst 0x454d1084  // ssublb z4.h, z4.b, z13.b\n"
    ".inst 0x448042cb  // smlalb z11.s, p4/M, z22.h, z0.h\n"
    ".inst 0x448046d0  // smlalt z16.s, p4/M, z22.h, z0.h\n"
    "ld1sb { z22.h }, p3/Z, [x11, x2]\n"
    ".inst 0x455112d6  // ssublb z22.h, z22.b, z17.b\n"
    ".inst 0x44804333  // smlalb z19.s, p4/M, z25.h, z0.h\n"
    ".inst 0x44804729  // smlalt z9.s, p4/M, z25.h, z0.h\n"
    ".inst 0x448043e7  // smlalb z7.s, p4/M, z31.h, z0.h\n"
    ".inst 0x448047e6  // smlalt z6.s, p4/M, z31.h, z0.h\n"
    ".inst 0x448043cc  // smlalb z12.s, p4/M, z30.h, z0.h\n"
    ".inst 0x448047c8  // smlalt z8.s, p4/M, z30.h, z0.h\n"
    "ld1sb { z0.h }, p4/Z, [x1, #7, MUL VL]\n"
    "inch x1, ALL, MUL #8\n"
    ".inst 0x4481432b  // smlalb z11.s, p4/M, z25.h, z1.h\n"
    ".inst 0x454d1000  // ssublb z0.h, z0.b, z13.b\n"
    ".inst 0x44814730  // smlalt z16.s, p4/M, z25.h, z1.h\n"
    "ld1sb { z25.h }, p3/Z, [x13, x2]\n"
    ".inst 0x44814313  // smlalb z19.s, p4/M, z24.h, z1.h\n"
    ".inst 0x45511339  // ssublb z25.h, z25.b, z17.b\n"
    ".inst 0x44814709  // smlalt z9.s, p4/M, z24.h, z1.h\n"
    ".inst 0x448143c7  // smlalb z7.s, p4/M, z30.h, z1.h\n"
    ".inst 0x448147c6  // smlalt z6.s, p4/M, z30.h, z1.h\n"
    ".inst 0x4481434c  // smlalb z12.s, p4/M, z26.h, z1.h\n"
    ".inst 0x44814748  // smlalt z8.s, p4/M, z26.h, z1.h\n"
    "ld1sb { z1.h }, p4/Z, [x1]\n"
    ".inst 0x454d1021  // ssublb z1.h, z1.b, z13.b\n"
    ".inst 0x4482430b  // smlalb z11.s, p4/M, z24.h, z2.h\n"
    ".inst 0x44824710  // smlalt z16.s, p4/M, z24.h, z2.h\n"
    "ld1sb { z24.h }, p3/Z, [x12, x2]\n"
    ".inst 0x45511318  // ssublb z24.h, z24.b, z17.b\n"
    ".inst 0x44824373  // smlalb z19.s, p4/M, z27.h, z2.h\n"
    ".inst 0x44824769  // smlalt z9.s, p4/M, z27.h, z2.h\n"
    ".inst 0x44824347  // smlalb z7.s, p4/M, z26.h, z2.h\n"
    ".inst 0x44824746  // smlalt z6.s, p4/M, z26.h, z2.h\n"
    ".inst 0x4482432c  // smlalb z12.s, p4/M, z25.h, z2.h\n"
    ".inst 0x44824728  // smlalt z8.s, p4/M, z25.h, z2.h\n"
    "ld1sb { z2.h }, p4/Z, [x1, #1, MUL VL]\n"
    ".inst 0x454d1042  // ssublb z2.h, z2.b, z13.b\n"
    ".inst 0x4483436b  // smlalb z11.s, p4/M, z27.h, z3.h\n"
    ".inst 0x44834770  // smlalt z16.s, p4/M, z27.h, z3.h\n"
    "ld1sb { z27.h }, p3/Z, [x10, x2]\n"
    ".inst 0x4551137b  // ssublb z27.h, z27.b, z17.b\n"
    ".inst 0x448342f3  // smlalb z19.s, p4/M, z23.h, z3.h\n"
    ".inst 0x448346e9  // smlalt z9.s, p4/M, z23.h, z3.h\n"
    ".inst 0x44834327  // smlalb z7.s, p4/M, z25.h, z3.h\n"
    ".inst 0x44834726  // smlalt z6.s, p4/M, z25.h, z3.h\n"
    ".inst 0x4483430c  // smlalb z12.s, p4/M, z24.h, z3.h\n"
    ".inst 0x44834708  // smlalt z8.s, p4/M, z24.h, z3.h\n"
    "ld1sb { z3.h }, p4/Z, [x1, #2, MUL VL]\n"
    ".inst 0x454d1063  // ssublb z3.h, z3.b, z13.b\n"
    ".inst 0x448442eb  // smlalb z11.s, p4/M, z23.h, z4.h\n"
    ".inst 0x448446f0  // smlalt z16.s, p4/M, z23.h, z4.h\n"
    "ld1sb { z23.h }, p3/Z, [x9, x2]\n"
    ".inst 0x455112f7  // ssublb z23.h, z23.b, z17.b\n"
    ".inst 0x44844393  // smlalb z19.s, p4/M, z28.h, z4.h\n"
    ".inst 0x44844789  // smlalt z9.s, p4/M, z28.h, z4.h\n"
    "ld1sb { z28.h }, p3/Z, [x26, x2]\n"
    ".inst 0x4551139c  // ssublb z28.h, z28.b, z17.b\n"
    ".inst 0x44844307  // smlalb z7.s, p4/M, z24.h, z4.h\n"
    ".inst 0x44844706  // smlalt z6.s, p4/M, z24.h, z4.h\n"
    ".inst 0x448442cc  // smlalb z12.s, p4/M, z22.h, z4.h\n"
    ".inst 0x448446c8  // smlalt z8.s, p4/M, z22.h, z4.h\n"
    "ld1sb { z4.h }, p4/Z, [x1, #3, MUL VL]\n"
    ".inst 0x454d1084  // ssublb z4.h, z4.b, z13.b\n"
    ".inst 0x448043eb  // smlalb z11.s, p4/M, z31.h, z0.h\n"
    ".inst 0x448047f0  // smlalt z16.s, p4/M, z31.h, z0.h\n"
    "ld1sb { z31.h }, p3/Z, [x28, x2]\n"
    ".inst 0x455113ff  // ssublb z31.h, z31.b, z17.b\n"
    ".inst 0x448043d3  // smlalb z19.s, p4/M, z30.h, z0.h\n"
    ".inst 0x448047c9  // smlalt z9.s, p4/M, z30.h, z0.h\n"
    ".inst 0x44804367  // smlalb z7.s, p4/M, z27.h, z0.h\n"
    ".inst 0x44804766  // smlalt z6.s, p4/M, z27.h, z0.h\n"
    ".inst 0x448042ec  // smlalb z12.s, p4/M, z23.h, z0.h\n"
    ".inst 0x448046e8  // smlalt z8.s, p4/M, z23.h, z0.h\n"
    "ld1sb { z0.h }, p4/Z, [x1, #4, MUL VL]\n"
    ".inst 0x454d1000  // ssublb z0.h, z0.b, z13.b\n"
    ".inst 0x448143cb  // smlalb z11.s, p4/M, z30.h, z1.h\n"
    ".inst 0x448147d0  // smlalt z16.s, p4/M, z30.h, z1.h\n"
    "ld1sb { z30.h }, p3/Z, [x27, x2]\n"
    ".inst 0x455113de  // ssublb z30.h, z30.b, z17.b\n"
    ".inst 0x44814353  // smlalb z19.s, p4/M, z26.h, z1.h\n"
    ".inst 0x44814749  // smlalt z9.s, p4/M, z26.h, z1.h\n"
    ".inst 0x448142e7  // smlalb z7.s, p4/M, z23.h, z1.h\n"
    ".inst 0x448146e6  // smlalt z6.s, p4/M, z23.h, z1.h\n"
    ".inst 0x448143ec  // smlalb z12.s, p4/M, z31.h, z1.h\n"
    ".inst 0x448147e8  // smlalt z8.s, p4/M, z31.h, z1.h\n"
    "ld1sb { z1.h }, p4/Z, [x1, #5, MUL VL]\n"
    ".inst 0x454d1021  // ssublb z1.h, z1.b, z13.b\n"
    ".inst 0x4482434b  // smlalb z11.s, p4/M, z26.h, z2.h\n"
    ".inst 0x44824750  // smlalt z16.s, p4/M, z26.h, z2.h\n"
    "ld1sb { z26.h }, p3/Z, [x25, x2]\n"
    ".inst 0x4551135a  // ssublb z26.h, z26.b, z17.b\n"
    ".inst 0x44824333  // smlalb z19.s, p4/M, z25.h, z2.h\n"
    ".inst 0x44824729  // smlalt z9.s, p4/M, z25.h, z2.h\n"
    ".inst 0x448243e7  // smlalb z7.s, p4/M, z31.h, z2.h\n"
    ".inst 0x448247e6  // smlalt z6.s, p4/M, z31.h, z2.h\n"
    ".inst 0x448243cc  // smlalb z12.s, p4/M, z30.h, z2.h\n"
    ".inst 0x448247c8  // smlalt z8.s, p4/M, z30.h, z2.h\n"
    "ld1sb { z2.h }, p4/Z, [x1, #6, MUL VL]\n"
    ".inst 0x454d1042  // ssublb z2.h, z2.b, z13.b\n"
    ".inst 0x4483432b  // smlalb z11.s, p4/M, z25.h, z3.h\n"
    ".inst 0x44834730  // smlalt z16.s, p4/M, z25.h, z3.h\n"
    "ld1sb { z25.h }, p3/Z, [x24, x2]\n"
    ".inst 0x45511339  // ssublb z25.h, z25.b, z17.b\n"
    ".inst 0x44834313  // smlalb z19.s, p4/M, z24.h, z3.h\n"
    ".inst 0x44834709  // smlalt z9.s, p4/M, z24.h, z3.h\n"
    ".inst 0x448343c7  // smlalb z7.s, p4/M, z30.h, z3.h\n"
    ".inst 0x448347c6  // smlalt z6.s, p4/M, z30.h, z3.h\n"
    ".inst 0x4483438c  // smlalb z12.s, p4/M, z28.h, z3.h\n"
    ".inst 0x44834788  // smlalt z8.s, p4/M, z28.h, z3.h\n"
    "ld1sb { z3.h }, p4/Z, [x1, #7, MUL VL]\n"
    "inch x1, ALL, MUL #8\n"
    ".inst 0x4484430b  // smlalb z11.s, p4/M, z24.h, z4.h\n"
    ".inst 0x454d1063  // ssublb z3.h, z3.b, z13.b\n"
    ".inst 0x44844710  // smlalt z16.s, p4/M, z24.h, z4.h\n"
    "ld1sb { z24.h }, p3/Z, [x23, x2]\n"
    ".inst 0x448442d3  // smlalb z19.s, p4/M, z22.h, z4.h\n"
    ".inst 0x45511318  // ssublb z24.h, z24.b, z17.b\n"
    ".inst 0x448446c9  // smlalt z9.s, p4/M, z22.h, z4.h\n"
    ".inst 0x44844387  // smlalb z7.s, p4/M, z28.h, z4.h\n"
    ".inst 0x44844786  // smlalt z6.s, p4/M, z28.h, z4.h\n"
    ".inst 0x4484434c  // smlalb z12.s, p4/M, z26.h, z4.h\n"
    ".inst 0x44844748  // smlalt z8.s, p4/M, z26.h, z4.h\n"
    "ld1sb { z4.h }, p4/Z, [x1]\n"
    "inch x1\n"
    ".inst 0x4480436b  // smlalb z11.s, p4/M, z27.h, z0.h\n"
    ".inst 0x454d1084  // ssublb z4.h, z4.b, z13.b\n"
    ".inst 0x44804770  // smlalt z16.s, p4/M, z27.h, z0.h\n"
    "ld1sb { z27.h }, p3/Z, [x22, x2]\n"
    ".inst 0x448042f3  // smlalb z19.s, p4/M, z23.h, z0.h\n"
    ".inst 0x4551137b  // ssublb z27.h, z27.b, z17.b\n"
    ".inst 0x448046e9  // smlalt z9.s, p4/M, z23.h, z0.h\n"
    ".inst 0x44804327  // smlalb z7.s, p4/M, z25.h, z0.h\n"
    ".inst 0x44804726  // smlalt z6.s, p4/M, z25.h, z0.h\n"
    "ld1sb { z25.h }, p3/Z, [x21, x2]\n"
    ".inst 0x45511339  // ssublb z25.h, z25.b, z17.b\n"
    ".inst 0x4480430c  // smlalb z12.s, p4/M, z24.h, z0.h\n"
    ".inst 0x44804708  // smlalt z8.s, p4/M, z24.h, z0.h\n"
    ".inst 0x448142eb  // smlalb z11.s, p4/M, z23.h, z1.h\n"
    ".inst 0x448146f0  // smlalt z16.s, p4/M, z23.h, z1.h\n"
    ".inst 0x448143f3  // smlalb z19.s, p4/M, z31.h, z1.h\n"
    ".inst 0x448147e9  // smlalt z9.s, p4/M, z31.h, z1.h\n"
    ".inst 0x44814307  // smlalb z7.s, p4/M, z24.h, z1.h\n"
    ".inst 0x44814706  // smlalt z6.s, p4/M, z24.h, z1.h\n"
    "ld1sb { z24.h }, p3/Z, [x20, x2]\n"
    ".inst 0x45511318  // ssublb z24.h, z24.b, z17.b\n"
    ".inst 0x4481436c  // smlalb z12.s, p4/M, z27.h, z1.h\n"
    ".inst 0x44814768  // smlalt z8.s, p4/M, z27.h, z1.h\n"
    ".inst 0x448243eb  // smlalb z11.s, p4/M, z31.h, z2.h\n"
    ".inst 0x448247f0  // smlalt z16.s, p4/M, z31.h, z2.h\n"
    ".inst 0x448243d3  // smlalb z19.s, p4/M, z30.h, z2.h\n"
    ".inst 0x448247c9  // smlalt z9.s, p4/M, z30.h, z2.h\n"
    ".inst 0x44824367  // smlalb z7.s, p4/M, z27.h, z2.h\n"
    ".inst 0x44824766  // smlalt z6.s, p4/M, z27.h, z2.h\n"
    "ld1sb { z27.h }, p3/Z, [x19, x2]\n"
    "inch x2\n"
    ".inst 0x4482432c  // smlalb z12.s, p4/M, z25.h, z2.h\n"
    "whilelt p2.s, x2, x0\n"
    ".inst 0x44824728  // smlalt z8.s, p4/M, z25.h, z2.h\n"
    "mov x19, x2\n"
    ".inst 0x448343cb  // smlalb z11.s, p4/M, z30.h, z3.h\n"
    ".inst 0x4551137b  // ssublb z27.h, z27.b, z17.b\n"
    ".inst 0x448347d0  // smlalt z16.s, p4/M, z30.h, z3.h\n"
    "incw x19\n"
    ".inst 0x44834393  // smlalb z19.s, p4/M, z28.h, z3.h\n"
    "whilelt p1.s, x19, x0\n"
    ".inst 0x44834789  // smlalt z9.s, p4/M, z28.h, z3.h\n"
    "whilelt p3.h, x2, x0\n"
    ".inst 0x44834327  // smlalb z7.s, p4/M, z25.h, z3.h\n"
    ".inst 0x44834726  // smlalt z6.s, p4/M, z25.h, z3.h\n"
    ".inst 0x4483430c  // smlalb z12.s, p4/M, z24.h, z3.h\n"
    ".inst 0x44834708  // smlalt z8.s, p4/M, z24.h, z3.h\n"
    ".inst 0x4484438b  // smlalb z11.s, p4/M, z28.h, z4.h\n"
    ".inst 0x44844790  // smlalt z16.s, p4/M, z28.h, z4.h\n"
    ".inst 0x44844353  // smlalb z19.s, p4/M, z26.h, z4.h\n"
    ".inst 0x44844749  // smlalt z9.s, p4/M, z26.h, z4.h\n"
    ".inst 0x04b5756b  // sqrdmulh z11.s, z11.s, z21.s\n"
    ".inst 0x04aa7610  // sqrdmulh z16.s, z16.s, z10.s\n"
    ".inst 0x04b57673  // sqrdmulh z19.s, z19.s, z21.s\n"
    ".inst 0x04aa7529  // sqrdmulh z9.s, z9.s, z10.s\n"
    "and z31.d, z11.d, z29.d\n"
    "asr z31.s, z31.s, #0x1f\n"
    "and z23.d, z16.d, z20.d\n"
    "and z25.d, z19.d, z29.d\n"
    "asr z23.s, z23.s, #0x1f\n"
    "and z18.d, z9.d, z20.d\n"
    ".inst 0x44844307  // smlalb z7.s, p4/M, z24.h, z4.h\n"
    "asr z25.s, z25.s, #0x1f\n"
    ".inst 0x44844706  // smlalt z6.s, p4/M, z24.h, z4.h\n"
    "asr z18.s, z18.s, #0x1f\n"
    "sqadd z11.s, z11.s, z31.s\n"
    ".inst 0x4484436c  // smlalb z12.s, p4/M, z27.h, z4.h\n"
    ".inst 0x04b574e7  // sqrdmulh z7.s, z7.s, z21.s\n"
    "sqadd z16.s, z16.s, z23.s\n"
    "sqadd z19.s, z19.s, z25.s\n"
    ".inst 0x04aa74c6  // sqrdmulh z6.s, z6.s, z10.s\n"
    "sqadd z9.s, z9.s, z18.s\n"
    "and z1.d, z7.d, z29.d\n"
    "asr z1.s, z1.s, #0x1f\n"
    "and z18.d, z6.d, z20.d\n"
    ".inst 0x04b5758c  // sqrdmulh z12.s, z12.s, z21.s\n"
    "asr z18.s, z18.s, #0x1f\n"
    ".inst 0x44844768  // smlalt z8.s, p4/M, z27.h, z4.h\n"
    ".inst 0x448293ab  // srshl z11.s, p4/M, z11.s, z29.s\n"
    "and z30.d, z12.d, z29.d\n"
    "asr z30.s, z30.s, #0x1f\n"
    "add z11.s, z11.s, z14.s\n"
    "sqadd z7.s, z7.s, z1.s\n"
    "sqadd z6.s, z6.s, z18.s\n"
    ".inst 0x04aa7508  // sqrdmulh z8.s, z8.s, z10.s\n"
    "smin z11.s, p4/M, z11.s, z15.s\n"
    ".inst 0x44829290  // srshl z16.s, p4/M, z16.s, z20.s\n"
    "sqadd z12.s, z12.s, z30.s\n"
    "and z3.d, z8.d, z20.d\n"
    "asr z3.s, z3.s, #0x1f\n"
    "add z16.s, z16.s, z14.s\n"
    "smax z11.s, p4/M, z11.s, z5.s\n"
    ".inst 0x448293b3  // srshl z19.s, p4/M, z19.s, z29.s\n"
    ".inst 0x44829289  // srshl z9.s, p4/M, z9.s, z20.s\n"
    "smin z16.s, p4/M, z16.s, z15.s\n"
    ".inst 0x448293a7  // srshl z7.s, p4/M, z7.s, z29.s\n"
    "add z19.s, z19.s, z14.s\n"
    "add z9.s, z9.s, z14.s\n"
    "sqadd z8.s, z8.s, z3.s\n"
    "add z7.s, z7.s, z14.s\n"
    "smax z16.s, p4/M, z16.s, z5.s\n"
    "smin z19.s, p4/M, z19.s, z15.s\n"
    "smin z9.s, p4/M, z9.s, z15.s\n"
    "smin z7.s, p4/M, z7.s, z15.s\n"
    "trn1 z11.h, z11.h, z16.h\n"
    "st1b { z11.h }, p0, [x7, x3]\n"
    "smax z19.s, p4/M, z19.s, z5.s\n"
    "smax z9.s, p4/M, z9.s, z5.s\n"
    "smax z7.s, p4/M, z7.s, z5.s\n"
    ".inst 0x44829286  // srshl z6.s, p4/M, z6.s, z20.s\n"
    ".inst 0x448293ac  // srshl z12.s, p4/M, z12.s, z29.s\n"
    "trn1 z19.h, z19.h, z9.h\n"
    "st1b { z19.h }, p0, [x8, x3]\n"
    "add z6.s, z6.s, z14.s\n"
    ".inst 0x44829288  // srshl z8.s, p4/M, z8.s, z20.s\n"
    "add z12.s, z12.s, z14.s\n"
    "smin z6.s, p4/M, z6.s, z15.s\n"
    "add z8.s, z8.s, z14.s\n"
    "smin z12.s, p4/M, z12.s, z15.s\n"
    "smax z6.s, p4/M, z6.s, z5.s\n"
    "smin z8.s, p4/M, z8.s, z15.s\n"
    "smax z12.s, p4/M, z12.s, z5.s\n"
    "trn1 z7.h, z7.h, z6.h\n"
    "st1b { z7.h }, p0, [x17, x3]\n"
    "smax z8.s, p4/M, z8.s, z5.s\n"
    "trn1 z12.h, z12.h, z8.h\n"
    "st1b { z12.h }, p0, [x16, x3]\n"
    "inch x3\n"
    "ldr x19, [%x[params], %[offsetof_Params_bias]]\n"
    "ld1w { z19.s }, p2/Z, [x19]\n"
    "ld1w { z6.s }, p1/Z, [x19, #1, MUL VL]\n"
    "uzp1 z11.s, z19.s, z6.s\n"
    "addvl x19, x19, #2\n"
    "str x19, [%x[params], %[offsetof_Params_bias]]\n"
    "uzp2 z16.s, z19.s, z6.s\n"
    "mov z19.d, z11.d\n"
    "ld1sb { z0.h }, p4/Z, [x1]\n"
    ".inst 0x454d1000  // ssublb z0.h, z0.b, z13.b\n"
    "mov z9.d, z16.d\n"
    "ld1sb { z1.h }, p4/Z, [x1, #1, MUL VL]\n"
    "mov z7.d, z11.d\n"
    "ld1sb { z2.h }, p4/Z, [x1, #2, MUL VL]\n"
    ".inst 0x454d1021  // ssublb z1.h, z1.b, z13.b\n"
    "mov z6.d, z16.d\n"
    "ld1sb { z3.h }, p4/Z, [x1, #3, MUL VL]\n"
    "mov z12.d, z11.d\n"
    "ld1sb { z4.h }, p4/Z, [x1, #4, MUL VL]\n"
    ".inst 0x454d1042  // ssublb z2.h, z2.b, z13.b\n"
    "mov z8.d, z16.d\n"
    "ldp x28, x27, [x5, #0x0]\n"
    ".inst 0x454d1063  // ssublb z3.h, z3.b, z13.b\n"
    "ldp x26, x25, [x5, #0x10]\n"
    ".inst 0x454d1084  // ssublb z4.h, z4.b, z13.b\n"
    "ldp x24, x23, [x5, #0x20]\n"
    "ldp x22, x21, [x5, #0x30]\n"
    "ldp x20, x19, [x5, #0x40]\n"
    "ld1sb { z31.h }, p3/Z, [x28, x2]\n"
    ".inst 0x455113ff  // ssublb z31.h, z31.b, z17.b\n"
    "ld1sb { z30.h }, p3/Z, [x27, x2]\n"
    "ld1sb { z29.h }, p3/Z, [x26, x2]\n"
    ".inst 0x455113de  // ssublb z30.h, z30.b, z17.b\n"
    "ld1sb { z28.h }, p3/Z, [x25, x2]\n"
    "ld1sb { z27.h }, p3/Z, [x24, x2]\n"
    ".inst 0x455113bd  // ssublb z29.h, z29.b, z17.b\n"
    "ld1sb { z23.h }, p3/Z, [x23, x2]\n"
    ".inst 0x4551139c  // ssublb z28.h, z28.b, z17.b\n"
    "ld1sb { z25.h }, p3/Z, [x22, x2]\n"
    "ld1sb { z24.h }, p3/Z, [x21, x2]\n"
    ".inst 0x4551137b  // ssublb z27.h, z27.b, z17.b\n"
    "ld1sb { z26.h }, p3/Z, [x20, x2]\n"
    ".inst 0x455112f7  // ssublb z23.h, z23.b, z17.b\n"
    "ld1sb { z22.h }, p3/Z, [x19, x2]\n"
    ".inst 0x45511339  // ssublb z25.h, z25.b, z17.b\n"
    ".inst 0x45511318  // ssublb z24.h, z24.b, z17.b\n"
    ".inst 0x4551135a  // ssublb z26.h, z26.b, z17.b\n"
    ".inst 0x455112d6  // ssublb z22.h, z22.b, z17.b\n"
    "b.any 1b\n"
    :
    : [offsetof_Params_bias] "I" (offsetof(Params, bias)), [offsetof_Params_inptrs] "I" (offsetof(Params, inptrs)), [offsetof_Params_n_channels] "I" (offsetof(Params, n_channels)), [offsetof_Params_outptrs] "I" (offsetof(Params, outptrs)), [offsetof_Params_requant] "I" (offsetof(Params, requant)), [offsetof_Params_requant_muls] "I" (offsetof(Params, requant_muls)), [offsetof_Params_requant_shifts] "I" (offsetof(Params, requant_shifts)), [offsetof_Params_weights] "I" (offsetof(Params, weights)), [offsetof_Requantize32_a_offset] "I" (offsetof(arm_gemm::Requantize32, a_offset)), [offsetof_Requantize32_b_offset] "I" (offsetof(arm_gemm::Requantize32, b_offset)), [offsetof_Requantize32_c_offset] "I" (offsetof(arm_gemm::Requantize32, c_offset)), [offsetof_Requantize32_maxval] "I" (offsetof(arm_gemm::Requantize32, maxval)), [offsetof_Requantize32_minval] "I" (offsetof(arm_gemm::Requantize32, minval)), [params] "r" (&params)
    : "cc", "memory", "p0", "p1", "p2", "p3", "p4", "x0", "x1", "x2", "x3", "x4", "x5", "x6", "x7", "x8", "x9", "x10", "x11", "x12", "x13", "x14", "x15", "x16", "x17", "x19", "x20", "x21", "x22", "x23", "x24", "x25", "x26", "x27", "x28", "z0", "z1", "z2", "z3", "z4", "z5", "z6", "z7", "z8", "z9", "z10", "z11", "z12", "z13", "z14", "z15", "z16", "z17", "z18", "z19", "z20", "z21", "z22", "z23", "z24", "z25", "z26", "z27", "z28", "z29", "z30", "z31"
  );
}

}  // namespace depthwise
}  // namespace arm_conv

#endif  // defined(__aarch64__) && defined(ARM_COMPUTE_ENABLE_SVE)
